/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

//One way fix by Olav Haugan
//09.12.01

package ZEN_IDL.be;

import java.io.*;
import java.io.PrintWriter;
import java.util.*;
import ZEN_IDL.ast.*;
import ZEN_IDL.be.symbolTableUtils.*;
import ZEN_IDL.be.helperUtils.*;

public class SymbolTableVisitor extends AbstractVisitor
{
	//----------------------- MODULE --------------------------
	public Object visit(ASTmodule node, Object data)
	{
		VisitorState mState = (VisitorState) data;
		mState.setCurrentNodeType( IDLParserTreeConstants.JJTMODULE );
		data = node.childrenAccept(this, data);
		mState.stkPackages.pop();
		return data;
	}
	
	public Object visit(ASTidentifier node, Object data)
	{
		VisitorState mState = (VisitorState) data;
		String name = node.getFirstToken().toString();
		switch( mState.getCurrentNodeType() )
		{
		case IDLParserTreeConstants.JJTMODULE:
			mState.stkPackages.push( name );
			(new File( mState.getPackagePath() )).mkdirs();
			mState.setCurrentNodeType( -1 );
			break;
		}
		data = node.childrenAccept(this, data);
		return data;
	}

	//---------------------- CONSTANTS ------------------------
	public Object visit(ASTconst_dcl node, Object data)
	{
		VisitorState mState = (VisitorState) data;
		String variableName = node.jjtGetChild(1).getFirstToken().toString();
		try{
			SymbolTable.addSymbol( new Symbol( variableName , mState.getPackage() ,
				IDLParserTreeConstants.JJTCONST_DCL , node.getFirstToken() ) );
		}catch( InvalidIdentifierException e ){
			System.out.println( e );
			mState.setErrorOccured( true );
		}
		return data;
	}

	//-------------------- INTERFACE --------------------------

	//currently not taking care of interface forward declarations.
	//assuming that all interfaces are defined before being inherited.

	InterfaceInheritence iSpec;

	public Object visit(ASTinterface_dcl node, Object data)
  	{
		VisitorState mState = (VisitorState) data;
		iSpec = new InterfaceInheritence( mState.getPackageVector() );
		
		//init vectors to store inheritence , methods
		mState.initializeMethodsVector();
		System.out.println( "Methods vector init...." );

		//set interface name
		String variableName = node.jjtGetChild(0).jjtGetChild(0).getFirstToken().toString();

		System.out.println( "----INTERFACE: " +variableName+ "----" );


		//go get the methods and inheritence spec...
		data = node.childrenAccept(this, data);

		//complete methods vector by adding inherited methods...
		Vector inheritenceVector = iSpec.getInheritence();
		System.out.println( "Numparents: " + inheritenceVector.size() );
		for( int i=0 ; i<inheritenceVector.size() ; i++){
			System.out.println( "Parent class : " );
			System.out.print( 	
				iSpec.getPackage( (Vector) inheritenceVector.elementAt( i ) ));
			System.out.println( " ... " +
				iSpec.getInterfaceName(  
				(Vector) inheritenceVector.elementAt( i ) ));

			Vector temp = (Vector) inheritenceVector.elementAt( i );
			String parentName = iSpec.getInterfaceName( temp );
			String parentPackage = iSpec.getPackage( temp );
			Symbol parent = SymbolTable.getSymbol( parentName , parentPackage );
			Vector parentMethods = parent.getMethods();

			for( int j=0 ; j<parentMethods.size() ; j++ )
				mState.addInterfaceMethod(
					(InterfaceMethod) parentMethods.elementAt(j));
		}

		//now add to synbol table
		try{
			System.out.println( "Adding interface to symbol table: " + variableName + "," + mState.getPackage() );
			SymbolTable.addSymbol( new Symbol( variableName , mState.getPackage() ,
				IDLParserTreeConstants.JJTINTERFACE_DCL , node.getFirstToken(),
				iSpec , mState.getInterfaceMethodsVector() ,
				false ) );
		}catch( InvalidIdentifierException e ){
			System.out.println( e );
			mState.setErrorOccured( true );
		}
		return data;
  	}
	
	public Object visit(ASTscoped_name node, Object data)
        {
                VisitorState mState = (VisitorState) data;

                String firstToken = node.getFirstToken().toString();
                boolean absoluteScope = false;
                if( firstToken.trim().equals( "::" ) )
                {
                        absoluteScope = true;
                }

                int numChildren = node.jjtGetNumChildren();
                Vector scopedNameElements = new Vector();
                for( int i = 0 ; i < numChildren ; i++ ){
                        scopedNameElements.add(
				(Object) node.jjtGetChild(i).getFirstToken().toString() );
		}
		//update iSpec only if the name is an interface.
		//since all the children are being visited...methods and constants shouldnt
		//be added as inheritence...
		String name = InterfaceInheritence.getInterfaceName( scopedNameElements );
		String pack = InterfaceInheritence.getPackage( scopedNameElements );
		if( pack.trim().equals( "" ) )
			pack = mState.getPackage();
		try{
			System.out.println( "inheritence: check: " + name + "," + pack );	
			if( SymbolTable.getSymbol( name , pack ).isInterface )
			{
				System.out.println( "inheritence: found: " + scopedNameElements );	
				iSpec.addSpec( scopedNameElements , !absoluteScope );
			}
		}catch( NullPointerException npe ){
			//ignore...
			System.out.println( "Ignoring null pointer exception..line139. SymbolTableVisitor" );
		}
                return data;
        }

	
	public Object visit(ASTabstract_interface_dcl node, Object data)
  	{
		//ALMOST A COPY OF INTERFACE_DCL
		VisitorState mState = (VisitorState) data;
		iSpec = new InterfaceInheritence( mState.getPackageVector() );
		
		//init vectors to store inheritence , methods
		mState.initializeMethodsVector();

		//set interface name
		String variableName = node.jjtGetChild(0).jjtGetChild(0).getFirstToken().toString();

		System.out.println( "----ABSTARCT INTERFACE: " +variableName+ "----" );


		//go get the methods and inheritence spec...	
		data = node.childrenAccept(this, data);

		//complete methods vector by adding inherited methods...
		Vector inheritenceVector = iSpec.getInheritence();
		System.out.println( "Numparents: " + inheritenceVector.size() );
		for( int i=0 ; i<inheritenceVector.size() ; i++){
			System.out.println( "Parent class : " );
			System.out.print( 	
				iSpec.getPackage( (Vector) inheritenceVector.elementAt( i ) ));
			System.out.println( " ... " +
				iSpec.getInterfaceName(  
				(Vector) inheritenceVector.elementAt( i ) ));

			Vector temp = (Vector) inheritenceVector.elementAt( i );
			String parentName = iSpec.getInterfaceName( temp );
			String parentPackage = iSpec.getPackage( temp );
			Symbol parent = SymbolTable.getSymbol( parentName , parentPackage );
			Vector parentMethods = parent.getMethods();

			for( int j=0 ; j<parentMethods.size() ; j++ )
				mState.addInterfaceMethod(
					(InterfaceMethod) parentMethods.elementAt(j));
		}

		//now add to synbol table
		try{
			SymbolTable.addSymbol( new Symbol( variableName , mState.getPackage() ,
				IDLParserTreeConstants.JJTABSTRACT_INTERFACE_DCL , node.getFirstToken(),
				iSpec , mState.getInterfaceMethodsVector() ,
				true ) );
		}catch( InvalidIdentifierException e ){
			System.out.println( e );
			mState.setErrorOccured( true );
		}
		return data;
  	}


/*****************************************************************************************************
RELEVANT GRAMMAR
void op_dcl() :
{}
{
  [ op_attribute() ] op_type_spec() identifier() parameter_dcls() [ raises_expr() ] [ context_expr() ]
}
-- NOTE: context_expr() not handled yet.

-- NOTE: param_dcls() is zero or more param_dcls()
void param_dcl() :
{}
{
  param_attribute() param_type_spec() simple_declarator()
}
******************************************************************************************************/
//One way fix by Olav Haugan
//09.12.01
	public Object visit( ASTop_dcl node, Object data )
	{
        	VisitorState currentState = (VisitorState)data;
        	currentState.initializeCurrentMethod();
        	int num = 0;
					String tmp = node.jjtGetChild(num).getFirstToken().toString();
					if(tmp.equals("oneway"))
					{
						currentState.setCurrentMethodAsOneway();
						num++;
					}
        	String returnType = CodeGenX.getFullName( node.jjtGetChild(num) , " " );
        	num++;
					String methodName = node.jjtGetChild(num).getFirstToken().toString();
       		currentState.setCurrentMethodReturnType( returnType );
        	currentState.setCurrentMethodName( methodName );
        	data = node.childrenAccept( this, data );
        	currentState = (VisitorState)data;
        	currentState.addCurrentMethod();
        	return data; 
   	}
	
    	public Object visit( ASTop_attribute node, Object data )
    	{
        	VisitorState currentState = (VisitorState)data;
        	currentState.setCurrentMethodAsOneway(); 
        	return data;
    	}
    	public Object visit( ASTparam_dcl node, Object data )
    	{
        	VisitorState currentState = (VisitorState)data;
        	currentState.initializeCurrentParam();
        	String inOut = node.jjtGetChild(0).getFirstToken().toString();
        	String paramType = CodeGenX.getFullName( node.jjtGetChild(1) , " " );
        	String paramName = node.jjtGetChild(2).getFirstToken().toString();
        	currentState.setCurrentParamInOut( inOut );
        	currentState.setCurrentParamType( paramType );
        	currentState.setCurrentParamName( paramName );
        	currentState.addCurrentParamToCurrentMethod();
        	return data;
    	}
	public Object visit( ASTraises_expr node , Object data )
	{
		return (new SymbolTableExceptionVisitor()).visit( node , data );
	}
	//---------------------- ENUMS ----------------------------
	public Object visit(ASTenum_type node, Object data)
	{
		VisitorState mState = (VisitorState) data;
		String variableName = node.jjtGetChild(0).getFirstToken().toString();
		try{
			SymbolTable.addSymbol( new Symbol( variableName , mState.getPackage() , 
				IDLParserTreeConstants.JJTENUM_TYPE , node.getFirstToken() ) );
		}catch( InvalidIdentifierException e ){
			System.out.println( e );
			mState.setErrorOccured( true );
		}
		data = node.childrenAccept(this, data);
		return data;
	}
	//----------------------------------- TYPEDEF -------------------------------------
        public Object visit(ASTtype_declarator node, Object data )
	{
		data = node.childrenAccept(new XTypedefVisitor(), data);
		VisitorState mState = (VisitorState) data;
                String variableName = mState.getTypedefDeclarator().toString();
		try{
			SymbolTable.addSymbol( new Symbol( variableName , mState.getPackage() , 
				IDLParserTreeConstants.JJTSTRUCT_TYPE , node.getFirstToken()) );
		}catch( InvalidIdentifierException e ){
			System.out.println( e );
			mState.setErrorOccured( true );
		}
		return data;
	}
		
	//----------------------------------- STRUCT --------------------------------------

	public Object visit(ASTstruct_type node, Object data)
	{	
		VisitorState mState = (VisitorState) data;
                String variableName = node.jjtGetChild(0).getFirstToken().toString();
		try{
			SymbolTable.addSymbol( new Symbol( variableName , mState.getPackage() , 
				IDLParserTreeConstants.JJTSTRUCT_TYPE , node.getFirstToken()) );
		}catch( InvalidIdentifierException e ){
			System.out.println( e );
			mState.setErrorOccured( true );
		}
		data = node.childrenAccept(this, data);
		return data;
	}

	//----------------------------------- UNION ----------------------------------------
        public Object visit(ASTunion_type node, Object data)
	{
		VisitorState mState = (VisitorState)data;
		String variableName = node.jjtGetChild(0).getFirstToken().toString();
		try{
			SymbolTable.addSymbol( new Symbol( variableName , mState.getPackage() , 
				IDLParserTreeConstants.JJTSTRUCT_TYPE , node.getFirstToken()) );
		}catch( InvalidIdentifierException e ){
			System.out.println( e );
			mState.setErrorOccured( true );
		}
		data = node.childrenAccept(this, data);
		return data;
        }

	//-------------------------------- EXCEPTION ----------------------------------------
	public Object visit(ASTexcept_dcl node, Object data)
	{
		VisitorState mState = (VisitorState)data;
		String variableName = node.jjtGetChild(0).getFirstToken().toString();
		try{
			SymbolTable.addSymbol( new Symbol( variableName, mState.getPackage() ,
				IDLParserTreeConstants.JJTEXCEPT_DCL , node.getFirstToken()) );
		}catch( InvalidIdentifierException e ){
			System.out.println( e );
			mState.setErrorOccured( true );
		}
		data = node.childrenAccept(this, data);
		return data;
	}

/***************************************************************************************************
	Attributes:

	attr_dcl:
		 [ "readonly" ] "attribute" param_type_spec() simple_declarator() ( "," simple_declarator() )*

	simple_declarator:
		identifier
***************************************************************************************************/
	
	public Object visit(ASTattr_dcl node, Object data)
	{
		System.out.println( "Processing attributes.." );

		VisitorState mState = (VisitorState)data;
		boolean readonly=false;
		int numChildren = node.jjtGetNumChildren();
		System.out.println( "numchil: "  + numChildren );
		int i=0;
		 
		
		if( node.getFirstToken().toString().trim().equalsIgnoreCase( "readonly" ) )
		{
			readonly=true;
		}

		String IDLType = node.jjtGetChild(i++).getFirstToken().toString().trim();
		for( ; i<numChildren ; i++ )
		{
			String attributeName = node.jjtGetChild(i).getFirstToken().toString().trim();
			try{
			SymbolTable.addSymbol( new Symbol( attributeName, mState.getPackage() ,
				IDLParserTreeConstants.JJTATTR_DCL , node.getFirstToken()) );
			}catch( InvalidIdentifierException e ){
				System.out.println( e );
				mState.setErrorOccured( true );
			}

			//add the currosponding

			//get method
       		 	mState.initializeCurrentMethod();
        		String returnType = IDLType;
        		String methodName = attributeName;
       			mState.setCurrentMethodReturnType( returnType );
	        	mState.setCurrentMethodName( methodName );
	        	mState.setCurrentMethodGIOPName( "_get_" + methodName );
        		mState.addCurrentMethod();

			if( !readonly )
			{
				//set method
       			 	mState.initializeCurrentMethod();
        			returnType = "void";
        			methodName = attributeName;
       				mState.setCurrentMethodReturnType( returnType );
	        		mState.setCurrentMethodName( methodName );

	        		mState.initializeCurrentParam();
       			 	String inOut = "in";
        			String paramType = IDLType; 
        			String paramName = "value";
        			mState.setCurrentParamInOut( inOut );
        			mState.setCurrentParamType( paramType );
        			mState.setCurrentParamName( paramName );
	        		mState.setCurrentMethodGIOPName( "_set_" + methodName );
        			mState.addCurrentParamToCurrentMethod();

        			mState.addCurrentMethod();
			}
		}
		return mState;
	}
}
