/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package ZEN_IDL.be.symbolTableUtils;

import java.util.*;
import ZEN_IDL.ast.*;


public class SymbolTable
{
	private static java.util.Hashtable symbolTable=new Hashtable();

	public static void addSymbol( Symbol node )
		throws InvalidIdentifierException
	{
		node.setSafeName( getSafeName( node.getName() , node.getPackage() , node.getLocation() ) );
		symbolTable.put( (node.getPackage()+"."+node.getName()).toLowerCase().trim() , node );
		System.out.println( "Adding symbol:\n" + node );
	}
	
	public static Symbol getSymbol( String name , String pkg )
	{
		name = name.trim(); pkg = pkg.trim();
		return (Symbol) symbolTable.get( (pkg+"."+name).toLowerCase().trim() );
	}

	public static boolean symbolExists( String name , String pkg )
	{
		name = name.trim(); pkg = pkg.trim();
		return (symbolTable.get( (pkg+"."+name).toLowerCase().trim() ) != null );
	}

	//----------------------------------------------------------
	//               Variable name checking code
	//----------------------------------------------------------

	private static String reservedWords[] = { "abstract" , "double" , "long" , 
		"readonly" , "unsigned" , "any" , "enum" , "module" , "sequence" , 
		"attribute" , "union" , "exception" , "native" , "short" , "ValueBase" , 
		"boolean" , "factory" , "object" , "string" , "ValueType" , "case" ,
		"false" , "octet" , "struct" , "void" , "char" , "fixed" , "oneway" ,
		"supports" , "wchar" , "const" , "float" , "out" , "switch" , "wstring" ,
		"context" , "in" , "private" , "true" , "custom" , "inout" , "public" ,
		"truncatable" , "default" , "interface" , "raises" , "typedef" };
	
	public static String getSafeName( String name , String pkg , String location )
		throws InvalidIdentifierException
	{
		name = name.trim(); pkg = pkg.trim();

		for( int i=0 ; i<reservedWords.length ; i++ ){
			if( name.equalsIgnoreCase( reservedWords[i] ))
			{
				throw new InvalidIdentifierException( "Error at " + location + "\n" +
					"Invalid Identifier: " + name + "\n" +
					"Name conflicts with a IDL reserved word: " + reservedWords[i] );
			}		
		}

		if( symbolExists( name , pkg ) )
		{
			throw new InvalidIdentifierException( "Error at " + location + "\n" +
				"Invalid Identifier: " + name + "\n" +
				"Name conflicts with a predefined symbol." );
		}

		String suffix="";
		if( name.length() > 6 ){
			suffix = name.substring( name.length() - 6 );
		}
		String safeName="";
		if( suffix.equalsIgnoreCase( "Helper" ) || suffix.equalsIgnoreCase( "Holder" ) )
		{
			safeName = "_" + name;
		}else{
			safeName = name;
		}

		return safeName;
	}
		
}
