/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

import mqc.*;
import mqc.holders.*;
import java.util.*;
import java.io.*;

public class Client
{
   public static void main(String[] args)
   {
      Object object;
      C_ORB orb = new C_ORB();
      System.out.println("ORB created!");

      String corbalocStr = System.getProperty("corbaloc");

      if (corbalocStr != null)
      {
         System.out.println("Using corbaloc property: " + corbalocStr);
         object = orb.corbaloc_to_object(corbalocStr);
      }
      else
      {
      System.out.println("Using args[0]: " + args[0]);
      object = orb.corbaloc_to_object(args[0]);
      }

      System.out.println("corbaloc to object!");

      banker.bank mybank = banker.bankHelper.narrow(object);
      System.out.println("Narrow!");

      prompt(mybank);
   }

   public static void printMenu()
   {
      System.out.println("--- BANK MENU ----------------------------------------------------");
      System.out.println("create [amount]             - Creates an Account");
      System.out.println("deposit [accntNumber] [amount]  - Deposits money in an Account");
      System.out.println("withdraw [accntNumber] [amount] - Withdraws money from an Account");
      System.out.println("balance [accntNumber]         - Checks the balance of an Account");
      System.out.println("? or help                  - Displays this menu");
      System.out.println("exit                     - Exits the program");
      System.out.println("------------------------------------------------------------------");
   }

   public static void prompt(banker.bank mybank)
   {
      String str;
      BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
      printMenu();
      while (true)
      {
         System.out.print("\n>");
         try
         {
            str = br.readLine();
            StringTokenizer tokenizer = new StringTokenizer(str, " ");
            String cmd = tokenizer.nextToken();
            if(cmd.equals("create"))
            {
               float amount = Float.parseFloat(tokenizer.nextToken());
               short acctNum = mybank.createAccount(amount);
               System.out.println("Account #" + acctNum + " Created.");
            }
            else if(cmd.equals("deposit"))
            {
               short acctNum = Short.parseShort(tokenizer.nextToken());
               float amount = Float.parseFloat(tokenizer.nextToken());
               if(mybank.deposit(acctNum, amount))
                  System.out.println("Depositited $" + amount + " to Account #" + acctNum);
            }
            else if(cmd.equals("withdraw"))
            {
               short acctNum = Short.parseShort(tokenizer.nextToken());
               float amount = Float.parseFloat(tokenizer.nextToken());
               if(mybank.withdraw(acctNum, amount))
                  System.out.println("Withdrew $" + amount + " from Account #" + acctNum);
            }
            else if(cmd.equals("balance"))
            {
               short acctNum = Short.parseShort(tokenizer.nextToken());
               float amount = mybank.checkBalance(acctNum);
               System.out.println("Balance of Account #" + acctNum + " is " + amount);
            }
            else if (cmd.equals("exit"))
            {
               System.exit(0);
            }
            else if (cmd.equals("?") || cmd.equals("help"))
            {
               printMenu();
            }
            else
            {
               System.out.println("Bad command format!");
               printMenu();
            }
         }
         catch (IOException ex)
         {
            System.out.println("Error reading from STDIN");
         }
         catch (NoSuchElementException ex)
         {
            System.out.println("Bad command format!");
            printMenu();
         }
      }
   }
}
