/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

include(`Config_macros.m4')

package mqc;

import mqc.ior.*;

import java.io.*;
import java.util.*;

/**
 * Client ORB
 * 
 * @author Olav Haugan
 * @author A. David McKinnon
 * @author Kim Christian Swenson
 */
public class C_ORB extends ORB
{
   /**
    * Converts an IOR to an object.
    *
    * @param str  IOR string
    * @return  an object
    */
   public Object string_to_object(String str)
   {
      IOR ior = new IOR(str);
      ior.decode();

      boolean bigEndian = !(Config.FORCE_ENDIANESS && !Config.BIG_ENDIAN);

      Delegate delegate;

      __GIOP_PROTOCOL__(     delegate = new GIOPDelegate( this, ior, bigEndian );     )
      __GIOPLITE_PROTOCOL__( delegate = new GIOPLiteDelegate( this, ior, bigEndian ); )
      __MQCIOP_PROTOCOL__(   delegate = new MQCIOPDelegate( this, ior, bigEndian );   )

      ObjImpl objectImpl = new ObjImpl();
      objectImpl.set_delegate(delegate);
      return objectImpl;
   }

   /**
    * Converts a corbaloc reference to an object.
    *
    * @param str  corbaloc string
    * @return  an object
    */
   public Object corbaloc_to_object(String str)
   {
      IOR ior = new IOR(null, str);
      ior.decodeCorbaloc();

      boolean bigEndian = !(Config.FORCE_ENDIANESS && !Config.BIG_ENDIAN);

      Delegate delegate;

      __GIOP_PROTOCOL__(     delegate = new GIOPDelegate( this, ior, bigEndian );     )
      __GIOPLITE_PROTOCOL__( delegate = new GIOPLiteDelegate( this, ior, bigEndian ); )
      __MQCIOP_PROTOCOL__(   delegate = new MQCIOPDelegate( this, ior, bigEndian );   )

      ObjImpl objectImpl = new ObjImpl();
      objectImpl.set_delegate(delegate);
      return objectImpl;
   }
}
