/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package mqc.groupcom;
/*
 * GroupCom.java
 *
 * Created on May 19, 2002, 4:48 AM
 */

import mqc.*;
import java.util.*;
import java.net.*;
import java.io.*;

/**
 *
 * @author  d3k076
 */
public class GroupCom
{

   // define some properties for the GroupCom object

   // define a string to hold the server name for this group
   //  member
   private String strMemberName;
   // define a string to hold the list of group members (server names and ports)
   private String strGroupNames;
   // define an integer to hold the number of the port that
   //  the GroupCom object will listen on
   private int intListenPort;
   // define a vector that will hold the members of the group
   //  (including system name (ip address) and listening port)
   //   and a listing of the last 50 messages that have been received
   private Vector vGroup;
   // define a vector that will act as a buffer for the messages
   //  that have been received
   private Vector vBuffer;
   // define a vector that will hold the status for sent messages (whether
   //  or not ack messages have been received once a message is sent)
   private Vector vStatus;
   // define a vector that will hold the status for join messages that are sent
   private Vector vJoinSent;
   // define a vector that will hold join messages that are recevied
   private Vector vJoinReceived;
   // define an integer that will represent the current message id
   private int intCurMessageID;
   // define an socket server object
   private SocketServer sockListener;
   // define a UDP socket server object
   private UDPSocketServer udpSockListener;
   // define an integer to holds the time out setting (in milliseconds)
   private int intTimeout;
   // define a string to hold the connection type
   private String strConType;
   // define a vector to hold the listeners for MessageEvents
   private transient Vector messageListeners;
   // define a string to indicate the type of ordering required
   private String strOrdering;
   // define a vector to act as a buffer for FIFO ordering
   private Vector vFIFOBuffer;
   // define a vector to hold the current received message id for FIFO buffering
   private Vector vFIFOMessageIDs;
   // define a vector to act as a buffer for Causal ordering
   private Vector vCausalBuffer;
   // define a vector to hold the current received message ids for Causal ordering
   private Vector vCausalMessageIDs;
   // define a string to hold the current sequence number for Total Ordering
   private String strTotalSequence;
   // define a Counter object to hold the current sequence number for total ordering
   private Counter cTotalSequence;
   // define an integer to hold the sequencer's number
   private int intSequencer = -1;
   // define a Counter object (to handle generating the sequence numbers)
   private Counter cSequencer;
   // define a newTotal object to handle the total ordering
   private newTotal tOrderer;
   // define a vector to hold the messages that arrive
   private Vector vNewMessages;
   // define a ProcessInput object to handle processing the messages
   private ProcessInput proIn;
   // define a newWatchMemberResponses object to handle monitoring receivers of messages
   private newWatchMemberResponses nWMR;
   // define a newMonitorSender object to handle monitoring senders of messages
   private newMonitorSender nMS;


   /** Creates a new instance of GroupCom */
   public GroupCom()
   {
      // this is the default constructor--initialize the listening port,
      //  the group vector, and the buffer
      this.intListenPort = -1;
      this.vGroup = new Vector();
      this.vBuffer = new Vector();
      // initialize the join buffers
      this.vJoinSent = new Vector();
      this.vJoinReceived = new Vector();
      // initialize the current message id
      this.intCurMessageID = 0;
      // initialize the member name
      // initialize the ordering to "none"
      this.strOrdering = "none";
      this.strMemberName = "";
      this.strGroupNames = "";
      this.vStatus = new Vector();
      this.strConType = "TCP";
      this.vNewMessages = new Vector();
      this.proIn = new ProcessInput();
      Thread tPI = new Thread(proIn);
      tPI.start();
      this.nWMR = new newWatchMemberResponses();
      this.nMS = new newMonitorSender();
      if(Config.GCTimeOut == 0)
      {
         intTimeout = 60000;
      }
      else
      {
         intTimeout = Config.GCTimeOut;
      }
   }

   /** Creates a new instance of GroupCom and initializes
    *  the listening port and group member list
    */
   public GroupCom(int intPort, String strGroupMembers, String strName, String strCType)
   {
      // this constructor allows the user to pass in the listening port and
      //  a list of the group members (format is <system name1>:<port1>,<system name2:<port2>)
      //  and the connection type (TCP or UDP)
      try
      {
         // assign the listenting port
         this.setListenPort(intPort);
         // initialize the ordering type
         this.strOrdering = "none";
         // process the string list of Group Members (load them into the vector)
         this.setGroup(strGroupMembers);
         // initialize the message buffer
         this.vBuffer = new Vector();
         // initialize the join buffers
         this.vJoinSent = new Vector();
         this.vJoinReceived = new Vector();
         // initialize the current message id
         this.intCurMessageID = 0;
         // set the member name
         this.setMemberName(strName);
         // initialize the status vector
         this.vStatus = new Vector();
         // set the connection type
         this.strConType = strCType;
         this.vNewMessages = new Vector();
         this.proIn = new ProcessInput();
         Thread tPI = new Thread(proIn);
         tPI.start();
         this.nWMR = new newWatchMemberResponses();
         this.nMS = new newMonitorSender();
         if(Config.GCTimeOut == 0)
         {
            intTimeout = 60000;
         }
         else
         {
            intTimeout = Config.GCTimeOut;
         }
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }

   /** Creates a new instance of GroupCom and initializes
    *  the listening port and group member list and ordering type
    */
   public GroupCom(int intPort, String strGroupMembers, String strName, String strCType, String strOrder)
   {
      // this constructor allows the user to pass in the listening port and
      //  a list of the group members (format is <system name1>:<port1>,<system name2:<port2>)
      //  and the connection type (TCP or UDP)
      try
      {
         // assign the listenting port
         this.setListenPort(intPort);
         // initialize the message buffer
         this.vBuffer = new Vector();
         // initialize the ordering buffers
         if(strOrder.equalsIgnoreCase("fifo"))
         {
            vFIFOMessageIDs = new Vector();
            vFIFOBuffer = new Vector();
         }
         else if(strOrder.equalsIgnoreCase("causal"))
         {
            vCausalMessageIDs = new Vector();
            vCausalBuffer = new Vector();
         }
         else if(strOrder.equalsIgnoreCase("total"))
         {
            strTotalSequence = "-1";
            cTotalSequence = new Counter();
            // here we can initialize the Counter object (to handle generating
            //  sequence numbers)
            cSequencer = new Counter();
            tOrderer = new newTotal();
         }
         // initialize the ordering type
         this.strOrdering = strOrder;
         // process the string list of Group Members (load them into the vector)
         this.setGroup(strGroupMembers);
         // initialize the join buffers
         this.vJoinSent = new Vector();
         this.vJoinReceived = new Vector();
         // initialize the current message id
         this.intCurMessageID = 0;
         // set the member name
         this.setMemberName(strName);
         // initialize the status vector
         this.vStatus = new Vector();
         // set the connection type
         this.strConType = strCType;
         this.vNewMessages = new Vector();
         this.proIn = new ProcessInput();
         Thread tPI = new Thread(proIn);
         tPI.start();
         this.nWMR = new newWatchMemberResponses();
         this.nMS = new newMonitorSender();
         if(Config.GCTimeOut == 0)
         {
            intTimeout = 60000;
         }
         else
         {
            intTimeout = Config.GCTimeOut;
         }
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }

   /** Add a method to get the listening port */
   public int getListenPort()
   {
      return this.intListenPort;
   }

   /** Add a method to set the listening port */
   public void setListenPort(int intPort)
   {
      this.intListenPort = intPort;
   }

   /** Add a method to set the timeout
    */
   public void setTimeout(int intTO)
   {
      this.intTimeout = intTO;
   }

   /** Add a method to get the member name
    */
   public String getMemberName()
   {
      return this.strMemberName;
   }

   /** Add a method to set the member name
    */
   public void setMemberName(String strName)
   {
      this.strMemberName = strName;
   }

   /** Add a method to get the list of group members */
   public Vector getGroup()
   {
      return this.vGroup;
   }

   /** add a method to set the list of group members
    *  (format is <system name1>:<port1>,<system name2:<port2>)
    */
   public void setGroup(String strGroupMembers)
   {
      try
      {
         // check to see if we are using Causal ordering
         if(strOrdering.equalsIgnoreCase("causal"))
         {
            //vCausalMessageIDs.clear();
            vCausalMessageIDs.removeAllElements();
         }
         // store the raw list of group members
         this.strGroupNames = strGroupMembers;
         // process the string list of Group Members (load them into the vector)
         vGroup = new Vector();
         StringTokenizer stk = new StringTokenizer(strGroupMembers,",");
         while(stk.hasMoreTokens())
         {
            String strCurDest = stk.nextToken();
            // now use another string tokenizer to bust up the name from
            //  the port
            StringTokenizer stkServer = new StringTokenizer(strCurDest,":");
            // extract the server name
            String strCurServerName = stkServer.nextToken();
            String strCurServerPort = stkServer.nextToken();
            // now add these values to the vectors
            Vector vTemp = new Vector();
            vTemp.addElement(strCurServerName);
            vTemp.addElement(strCurServerPort);
            // next we need to create an empty vector to add which will hold the
            //  id's of the last 50 messages that have been received from this
            //  sender
            Vector vHistory = new Vector();
            // add the history vector to the temp vector (which will be added to the group
            //  vector)
            vTemp.addElement(vHistory);
            // now add this vector to the group vector
            this.vGroup.addElement(vTemp);
            // next we need to see if we need to add this member to the vector timestamp
            //  vector (if the ordering is set to causal)
            if(strOrdering.equalsIgnoreCase("causal"))
            {
               // here we need to add the member and initialize its sequence number to -1
               Vector vTSTemp = new Vector();
               vTSTemp.addElement(strCurDest);
               vTSTemp.addElement("-1");
               vCausalMessageIDs.addElement(vTSTemp);
            }
         }
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }

   /** Add a method to add a group member
    */
   public void addGroupMember(String strNewMember, String strJoinMessageID)
   {
      try
      {
         // add the group member to the group list
         this.strGroupNames = this.strGroupNames + "," + strNewMember;
         // Add the new group member to vGroup
         StringTokenizer stkServer = new StringTokenizer(strNewMember,":");
         // extract the server name
         String strCurServerName = stkServer.nextToken();
         String strCurServerPort = stkServer.nextToken();
         // now add these values to the vectors
         Vector vTemp = new Vector();
         vTemp.addElement(strCurServerName);
         vTemp.addElement(strCurServerPort);
         // next we need to create an empty vector to add which will hold the
         //  id's of the last 50 messages that have been received from this
         //  sender
         Vector vHistory = new Vector();
         // add the history vector to the temp vector (which will be added to the group
         //  vector)
         vTemp.addElement(vHistory);
         // now add this vector to the group vector
         this.vGroup.addElement(vTemp);
         // next we need to see if we need to add this member to the vector timestamp
         //  vector (if the ordering is set to causal)
         if(strOrdering.equalsIgnoreCase("causal"))
         {
            synchronized (vCausalMessageIDs)
            {
               // here we need to add the member and initialize its sequence number to -1
               Vector vTSTemp = new Vector();
               vTSTemp.addElement(strNewMember);
               vTSTemp.addElement(strJoinMessageID);
               vCausalMessageIDs.addElement(vTSTemp);
            }
         }
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }


   /** Add a method to join a group
    */
   public void joinGroup(String strContactInfo)
   {
      try
      {
         // first we need to break up the contact info into the server name and port
         StringTokenizer stk = new StringTokenizer(strContactInfo,":");
         String strServerName = stk.nextToken();
         String strServerPort = stk.nextToken();
         int intServerPort = Integer.parseInt(strServerPort);
         // grab the message id for the join message and convert it to a string
         String strMessageID = Integer.toString(this.intCurMessageID);
         // define a vector to hold the join info
         Vector vJoinInfo = new Vector();
         // next we need to open a connection to the server--first check what the
         //  transport setting is
         if(strConType.equalsIgnoreCase("TCP"))
         {
            //System.out.println("Sending join message to " + strServerName + ":" + strServerPort + " via TCP");
            // here we need to make a TCP connection to the group member
            // open up a TCP socket to the group member
            Socket sockSendMessage = new Socket(strServerName,intServerPort);
            // next we need to build a string with the message
            String strMessage = strMemberName + ":" + intListenPort + "~new~join~" + strMessageID;
            byte[] msg = strMessage.getBytes();
            // now use a DataOutputStream to print the byte array to the socket
            DataOutputStream dosOut = new DataOutputStream(sockSendMessage.getOutputStream());
            dosOut.write(msg);
            dosOut.flush();
            dosOut.close();
            sockSendMessage.close();
            /*
            // next decorate the socket with a print writer
            PrintWriter pwSendMessage = new PrintWriter(sockSendMessage.getOutputStream());
            // next print out the message info to the socket
            pwSendMessage.println(strMemberName + ":" + intListenPort + "~new~join~" + strMessageID);
            // force a flush of the print writer
            pwSendMessage.flush();
            // close and deallocate the socket and print writer
            pwSendMessage.close();
            sockSendMessage.close();
             */
         }
         else
         {
            // here we need to make a UDP connection to the group member
            // get everything set up to transmit the message
            //System.out.println("Sending join message to " + strServerName + ":" + strServerPort + " via UDP");
            String strMessage = strMemberName + ":" + intListenPort + "~new~join~" + strMessageID;
            byte[] msg = strMessage.getBytes();
            // create a new datagram socket to send the message
            DatagramSocket ds = new DatagramSocket();
            InetAddress ia = InetAddress.getByName(strServerName);
            // create the outgoing data packet
            DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,intServerPort);
            // now send the packet to the server
            ds.send(outPacket);
            ds.close();
         }
         // add the info to the vJoinInfo vector
         vJoinInfo.addElement(strMessageID);
         vJoinInfo.addElement(strContactInfo);
         // add the message to the vJoinSent buffer
         synchronized (vJoinSent)
         {
            vJoinSent.addElement(vJoinInfo);
         }
         // increment the message id for this sender
         this.intCurMessageID++;
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }


   /** Add a method to search for a member of the group (if it returns true the member was found)
    */
   private boolean findMember(String strFMember)
   {
      boolean blnFound = false;
      // here the string passed in will contain the member server name and port
      // set up a loop to find the matching group member if they exist

      for(int i=0;i<vGroup.size();i++)
      {
         // first extract the current member from the vector
         Vector vTemp = (Vector) vGroup.elementAt(i);
         // next extract the server name and listening port
         String strCurServer = (String) vTemp.elementAt(0);
         String strPort = (String) vTemp.elementAt(1);
         // next build the combined string to check against
         String strCheck = strCurServer + ":" + strPort;
         // check to see if this matches the member to be removed
         if(strCheck.equalsIgnoreCase(strFMember))
         {
            // we have found the match--indicate the group member was found
            blnFound = true;
            // now break out of the loop
            break;
         }
      }
      return blnFound;
   }

   /** Add a method to remove a member from the group
    */
   private void removeMember(String strRMember)
   {
      // here the string passed in will contain the member server name and port
      // first we need to remove the member from vGroup
      for(int i=0;i<vGroup.size();i++)
      {
         // first extract the current member from the vector
         Vector vTemp = (Vector) vGroup.elementAt(i);
         // next extract the server name and listening port
         String strCurServer = (String) vTemp.elementAt(0);
         String strPort = (String) vTemp.elementAt(1);
         // next build the combined string to check against
         String strCheck = strCurServer + ":" + strPort;
         // check to see if this matches the member to be removed
         if(strCheck.equalsIgnoreCase(strRMember))
         {
            // we have found the match--remove this member from the vector
            System.out.println("Removing Member: " + strRMember);
            //synchronized(vGroup) {
            vGroup.removeElement(vTemp);
            //}
            // need to add in code to clean out vBuffer
            Vector vSearchBuffer = (Vector) vBuffer.clone();

            // Iterator itVSB = vSearchBuffer.iterator();
            Enumeration itVSB = vSearchBuffer.elements();
            //while(itVSB.hasNext()) {
            while(itVSB.hasMoreElements())
            {
               //Vector vSTemp = (Vector) itVSB.next();
               Vector vSTemp = (Vector) itVSB.nextElement();
               // extract the sender info and message id
               String strBufSender = (String) vSTemp.elementAt(0);
               if(strBufSender.equalsIgnoreCase(strRMember))
               {
                  vBuffer.removeElement(vSTemp);
               }
            }


            // now break out of the loop
            break;
         }
      }

      // next we can remove this member from the vTimestamp vector (if causal ordering is on)
      if(strOrdering.equalsIgnoreCase("causal"))
      {
         synchronized (vCausalMessageIDs)
         {
            // here we need to set up a loop to find the corresponding member and remove it
            for(int i=0;i<vCausalMessageIDs.size();i++)
            {
               // extract the current element in the vector
               Vector vCurCausal = (Vector) vCausalMessageIDs.elementAt(i);
               // extract the server info
               String strCurServerInfo = (String) vCurCausal.elementAt(0);
               // check to see if we have found a match
               if(strCurServerInfo.equalsIgnoreCase(strRMember))
               {
                  // we have found a match--remove this member from the vector
                  //synchronized(vCausalMessageIDs) {
                  vCausalMessageIDs.removeElement(vCurCausal);
                  //}
                  break;
               }
            }
         }
      }

      // next we can build a new group member listing string (based off the new version
      //  of vGroup
      String strGroupListTemp = "";

      for(int i=0;i<vGroup.size();i++)
      {
         // first extract the current member from the vector
         Vector vTemp = (Vector) vGroup.elementAt(i);
         // next extract the server name and listening port
         String strCurServer = (String) vTemp.elementAt(0);
         String strPort = (String) vTemp.elementAt(1);
         // next build the combined string to check against
         String strCurMember = strCurServer + ":" + strPort;
         // next add the member to the string
         if(strGroupListTemp.equalsIgnoreCase(""))
         {
            strGroupListTemp = strCurMember;
         }
         else
         {
            strGroupListTemp = strGroupListTemp + "," + strCurMember;
         }
      }

      // now replace the class variable
      this.strGroupNames = strGroupListTemp;
   }

   /** Add a method to build the vector timestamp for causal ordering
    */
   public String getVectorTimestamp()
   {
      String strVTimestamp = "";
      try
      {
         Vector vCMCopy;
         // make a clone of the current vCausalMessageIDs
         synchronized (vCausalMessageIDs)
         {
            vCMCopy = (Vector) vCausalMessageIDs.clone();
         }
         // next set up a loop to build the timestamp string (pipe and comma delimited)
         for(int i=0;i<vCMCopy.size();i++)
         {
            Vector vCurElem = (Vector) vCMCopy.elementAt(i);
            String strCurServer = (String) vCurElem.elementAt(0);
            String strCurMessageID = (String) vCurElem.elementAt(1);
            // add the elements to the string
            if(strCurServer.equalsIgnoreCase(this.strMemberName + ":" + Integer.toString(this.intListenPort)))
            {
               if(strVTimestamp.equalsIgnoreCase(""))
               {
                  strVTimestamp = strCurServer + "|" + Integer.toString(this.intCurMessageID);
               }
               else
               {
                  strVTimestamp = strVTimestamp + "," + strCurServer + "|" + Integer.toString(this.intCurMessageID);
               }
            }
            else
            {

               if(strVTimestamp.equalsIgnoreCase(""))
               {
                  strVTimestamp = strCurServer + "|" + strCurMessageID;
               }
               else
               {
                  strVTimestamp = strVTimestamp + "," + strCurServer + "|" + strCurMessageID;
               }
            }
         }
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
      return strVTimestamp;
   }

   /** Add a method to send a message to all the group members
    */
   public synchronized void sendMessage(String strMessageData,String strMType,String strOrigSenderInfo,String strOrigMessageID)
   {

      try
      {
         //System.out.println("Sending a message");
         // first we need to build the message vector that will hold the status
         Vector vTemp = new Vector();
         String strMessageID = Integer.toString(this.intCurMessageID);
         String strVectorTS = "";
         if(strOrdering.equalsIgnoreCase("causal"))
         {
            strVectorTS = strMessageID + "|" + getVectorTimestamp();
         }
         int intNextSequence = -1;
         // check to see if this is a sequence message
         if(strMType.equalsIgnoreCase("sequence"))
         {
            // here we need to get the next sequence number
            intNextSequence = cSequencer.incCounter();
         }

         Vector vSLoad = new Vector();
         if(strOrdering.equalsIgnoreCase("causal"))
         {
            vSLoad.addElement(strVectorTS);
            // here I need to send the message to causal and let it deal with updating the
            //  vector timestamp

            // here we can update vCausalMessageIDs for this message
            Vector vVUpdate = new Vector();
            String strMemberInfo = strMemberName + ":" + Integer.toString(intListenPort);
            vVUpdate.addElement(strMemberInfo);
            vVUpdate.addElement(Integer.toString(intCurMessageID));
            synchronized (vCausalMessageIDs)
            {
               for(int v=0;v<vCausalMessageIDs.size();v++)
               {
                  Vector vCurElem = (Vector) vCausalMessageIDs.elementAt(v);
                  String strCurHost = (String) vCurElem.elementAt(0);
                  if(strCurHost.equalsIgnoreCase(strMemberInfo))
                  {
                     // here we have found the element to update
                     vCausalMessageIDs.setElementAt(vVUpdate,vCausalMessageIDs.indexOf(vCurElem));
                     break;
                  }
               }
            }
         }
         else
         {
            vSLoad.addElement(Integer.toString(intCurMessageID));
         }
         vSLoad.addElement(vTemp);
         // see what vector we need to load the status into based on the message type
         if(strMType.equalsIgnoreCase("newMember") || strMType.equalsIgnoreCase("sfnm"))
         {
            // here we need to check and see if the group size is 1
            if(vGroup.size() == 1)
            {
               // here there is nothing to monitor
            }
            else
            {
               // add the status to the vJoinSent vector
               synchronized (vJoinSent)
               {
                  vJoinSent.addElement(vSLoad);
               }
               // here we need to kick off a thread to monitor the responses (and handle any
               //  time out issues)
               /*
               WatchMemberResponses wmr = new WatchMemberResponses("newMember");
               wmr.setMessageID(this.intCurMessageID);
               Thread tWMR = new Thread(wmr);
               // start the thread
               tWMR.start();
                */
            }
         }
         else
         {
            if(vGroup.size() == 1)
            {
               // here there is nothing to monitor
            }
            else
            {
               synchronized (vStatus)
               {
                  vStatus.addElement(vSLoad);
               }
               // here we need to kick off a thread to monitor the responses (and handle any
               //  time out issues)
               /*
               WatchMemberResponses wmr = new WatchMemberResponses();
               wmr.setMessageID(this.intCurMessageID);
               Thread tWMR = new Thread(wmr);
               // start the thread
               tWMR.start();
                */
            }
         }

         // next we need to set up a loop to open a socket and send the message
         //  to all group members (not including this member)
         //synchronized (vGroup) {
         for(int i=0;i<vGroup.size();i++)
         {
            // first extract the current member from the vector
            Vector vTempA = (Vector) vGroup.elementAt(i);
            // next extract the server name and listening port
            String strCurServer = (String) vTempA.elementAt(0);
            String strTemp = (String) vTempA.elementAt(1);
            int intCurListenPort = Integer.parseInt(strTemp);
            // check to see what type of message this is going to be
            if(strMType.equalsIgnoreCase("message"))
            {
               // here this is a regular message to be send (not a resend)
               // next check to see that this is not the same server as this one (don't
               if(strCurServer.equalsIgnoreCase(this.strMemberName) && intCurListenPort == this.intListenPort)
               {
                  // do nothing here (it is the local one)
               }
               else
               {
                  //System.out.println("Sending a message to " + strCurServer + ":" + intCurListenPort);
                  // add this member to the status vector
                  Vector vCurStatus = new Vector();
                  String strCurMember = strCurServer + ":" + intCurListenPort;
                  vCurStatus.addElement(strCurMember);
                  vCurStatus.addElement("false");
                  vTemp.addElement(vCurStatus);
                  // now we can send the message--determine if we are using TCP or UDP
                  if(strConType.equalsIgnoreCase("TCP"))
                  {
                     // here we can send the message--first open up a socket connection
                     Socket sockSendMessage = new Socket(strCurServer,intCurListenPort);
                     // next we need to build a string with the message
                     String strMessage;
                     if(strOrdering.equalsIgnoreCase("causal"))
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~message~" + strVectorTS + "~" + strMessageData;
                     }
                     else
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~message~" + strMessageID + "~" + strMessageData;
                     }
                     byte[] msg = strMessage.getBytes();
                     // now use a DataOutputStream to print the byte array to the socket
                     DataOutputStream dosOut = new DataOutputStream(sockSendMessage.getOutputStream());
                     dosOut.write(msg);
                     dosOut.flush();
                     dosOut.close();
                     sockSendMessage.close();
                     /*
                     // next decorate the socket with a print writer
                     PrintWriter pwSendMessage = new PrintWriter(sockSendMessage.getOutputStream());
                     // next print out the message info to the socket
                     if(strOrdering.equalsIgnoreCase("causal")) {
                         pwSendMessage.println(strMemberName + ":" + intListenPort + "~" + strGroupNames + "~message~" + strVectorTS + "~" + strMessageData);
                     } else {
                         pwSendMessage.println(strMemberName + ":" + intListenPort + "~" + strGroupNames + "~message~" + strMessageID + "~" + strMessageData);
                     }
                     // pwSendMessage.println(strMessageData);
                     // force a flush of the print writer
                     pwSendMessage.flush();
                     // close and deallocate the socket and print writer
                     pwSendMessage.close();
                     sockSendMessage.close();
                      */
                  }
                  else
                  {
                     // here we are using UDP--get everything set up to transmit the message
                     String strMessage;
                     if(strOrdering.equalsIgnoreCase("causal"))
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~message~" + strVectorTS + "~" + strMessageData;
                     }
                     else
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~message~" + strMessageID + "~" + strMessageData;
                     }
                     byte[] msg = strMessage.getBytes();
                     // create a new datagram socket to send the message
                     DatagramSocket ds = new DatagramSocket();
                     InetAddress ia = InetAddress.getByName(strCurServer);
                     // create the outgoing data packet
                     DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,intCurListenPort);
                     // now send the packet to the server
                     ds.send(outPacket);
                     ds.close();
                  }
               }
            }
            else if(strMType.equalsIgnoreCase("sequence"))
            {
               // here this is a sequence message to be sent
               // next check to see that this is not the server that sent the original message
               // String strCurServerInfo = strCurServer + ":" + Integer.toString(intCurListenPort);
               // get the next sequence number
               if(strCurServer.equalsIgnoreCase(this.strMemberName) && intCurListenPort == this.intListenPort)
               {
                  String strTOMessageID = "s|" + strOrigSenderInfo + "|" + strOrigMessageID + "|" + Integer.toString(intNextSequence);
                  this.fireReceiveMessage(strMemberName + ":" + Integer.toString(intListenPort),strGroupNames,strTOMessageID,"");
                  // do nothing here (it is the sender)
               }
               else
               {
                  //System.out.println("Sending a message to " + strCurServer + ":" + intCurListenPort);
                  // add this member to the status vector
                  Vector vCurStatus = new Vector();
                  String strCurMember = strCurServer + ":" + intCurListenPort;
                  vCurStatus.addElement(strCurMember);
                  vCurStatus.addElement("false");
                  vTemp.addElement(vCurStatus);
                  // now we can send the message--determine if we are using TCP or UDP
                  if(strConType.equalsIgnoreCase("TCP"))
                  {
                     // here we can send the message--first open up a socket connection
                     Socket sockSendMessage = new Socket(strCurServer,intCurListenPort);
                     String strTOMessageID = "s|" + strOrigSenderInfo + "|" + strOrigMessageID + "|" + Integer.toString(intNextSequence);
                     String strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sequence~" + strMessageID + "~" + strTOMessageID;
                     byte[] msg = strMessage.getBytes();
                     // now use a DataOutputStream to print the byte array to the socket
                     DataOutputStream dosOut = new DataOutputStream(sockSendMessage.getOutputStream());
                     dosOut.write(msg);
                     dosOut.flush();
                     dosOut.close();
                     sockSendMessage.close();
                     /*
                     // next decorate the socket with a print writer
                     PrintWriter pwSendMessage = new PrintWriter(sockSendMessage.getOutputStream());
                     String strTOMessageID = "s|" + strOrigSenderInfo + "|" + strOrigMessageID + "|" + Integer.toString(intNextSequence);
                     pwSendMessage.println(strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sequence~" + strMessageID + "~" + strTOMessageID);

                     // pwSendMessage.println(strMessageData);
                     // force a flush of the print writer
                     pwSendMessage.flush();
                     // close and deallocate the socket and print writer
                     pwSendMessage.close();
                     sockSendMessage.close();
                      */
                  }
                  else
                  {
                     // here we are using UDP--get everything set up to transmit the message
                     String strTOMessageID = "s|" + strOrigSenderInfo + "|" + strOrigMessageID + "|" + Integer.toString(intNextSequence);
                     String strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sequence~" + strMessageID + "~" + strTOMessageID;
                     byte[] msg = strMessage.getBytes();
                     // create a new datagram socket to send the message
                     DatagramSocket ds = new DatagramSocket();
                     InetAddress ia = InetAddress.getByName(strCurServer);
                     // create the outgoing data packet
                     DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,intCurListenPort);
                     // now send the packet to the server
                     ds.send(outPacket);
                     ds.close();
                  }
               }
            }
            else if(strMType.equalsIgnoreCase("sf"))
            {
               // here the message is a sf (sender failed message)
               // next check to see that this is not the same server as this one (don't
               if(strCurServer.equalsIgnoreCase(this.strMemberName) && intCurListenPort == this.intListenPort)
               {
                  // do nothing here (it is the local one)
               }
               else if(strOrigSenderInfo.equalsIgnoreCase(strCurServer + ":" + strTemp))
               {
                  // do nothing here (it is the failed sender)
               }
               else
               {
                  //System.out.println("Sending an sf message to " + strCurServer + ":" + intCurListenPort);
                  // add this member to the status vector
                  Vector vCurStatus = new Vector();
                  String strCurMember = strCurServer + ":" + intCurListenPort;
                  vCurStatus.addElement(strCurMember);
                  vCurStatus.addElement("false");
                  vTemp.addElement(vCurStatus);
                  // now we can send the message--determine if we are using TCP or UDP
                  if(strConType.equalsIgnoreCase("TCP"))
                  {
                     // here we can send the message--first open up a socket connection
                     Socket sockSendMessage = new Socket(strCurServer,intCurListenPort);
                     String strMessage;
                     if(strOrdering.equalsIgnoreCase("causal"))
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sf~" + strVectorTS + "~" + strOrigSenderInfo + "~" + strOrigMessageID + "~" + strMessageData;
                     }
                     else
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sf~" + strMessageID + "~" + strOrigSenderInfo + "~" + strOrigMessageID + "~" + strMessageData;
                     }
                     byte[] msg = strMessage.getBytes();
                     // now use a DataOutputStream to print the byte array to the socket
                     DataOutputStream dosOut = new DataOutputStream(sockSendMessage.getOutputStream());
                     dosOut.write(msg);
                     dosOut.flush();
                     dosOut.close();
                     sockSendMessage.close();
                     /*
                     // next decorate the socket with a print writer
                     PrintWriter pwSendMessage = new PrintWriter(sockSendMessage.getOutputStream());
                     // next print out the message info to the socket
                     if(strOrdering.equalsIgnoreCase("causal")) {
                         pwSendMessage.println(strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sf~" + strVectorTS + "~" + strOrigSenderInfo + "~" + strOrigMessageID + "~" + strMessageData);
                     } else {
                         pwSendMessage.println(strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sf~" + strMessageID + "~" + strOrigSenderInfo + "~" + strOrigMessageID + "~" + strMessageData);
                     }
                     // pwSendMessage.println(strMessageData);
                     // force a flush of the print writer
                     pwSendMessage.flush();
                     // close and deallocate the socket and print writer
                     pwSendMessage.close();
                     sockSendMessage.close();
                      */
                  }
                  else
                  {
                     // here we are using UDP--get everything set up to transmit the message
                     String strMessage;
                     if(strOrdering.equalsIgnoreCase("causal"))
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sf~" + strVectorTS + "~" + strOrigSenderInfo + "~" + strOrigMessageID + "~" + strMessageData;
                     }
                     else
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sf~" + strMessageID + "~" + strOrigSenderInfo + "~" + strOrigMessageID + "~" + strMessageData;
                     }
                     byte[] msg = strMessage.getBytes();
                     // create a new datagram socket to send the message
                     DatagramSocket ds = new DatagramSocket();
                     InetAddress ia = InetAddress.getByName(strCurServer);
                     // create the outgoing data packet
                     DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,intCurListenPort);
                     // now send the packet to the server
                     ds.send(outPacket);
                     ds.close();
                  }
               }
            }
            else if(strMType.equalsIgnoreCase("newMember"))
            {
               // here we need to send the newMember message
               // next check to see that this is not the same server as this one (don't
               if(strCurServer.equalsIgnoreCase(this.strMemberName) && intCurListenPort == this.intListenPort)
               {
                  // do nothing here (it is the local one)
               }
               else
               {
                  //System.out.println("Sending a newMember message to " + strCurServer + ":" + intCurListenPort);
                  // add this member to the status vector
                  Vector vCurStatus = new Vector();
                  String strCurMember = strCurServer + ":" + intCurListenPort;
                  StringTokenizer stkDataID = new StringTokenizer(strMessageData,"|");
                  String strNewMemberInfo = stkDataID.nextToken();
                  String strJoinMessageID = stkDataID.nextToken();
                  vCurStatus.addElement(strCurMember);
                  vCurStatus.addElement("false");
                  vCurStatus.addElement(strNewMemberInfo);
                  vCurStatus.addElement(strJoinMessageID);
                  vTemp.addElement(vCurStatus);
                  // now we can send the message--determine if we are using TCP or UDP
                  if(strConType.equalsIgnoreCase("TCP"))
                  {
                     // here we can send the message--first open up a socket connection
                     Socket sockSendMessage = new Socket(strCurServer,intCurListenPort);
                     String strMessage;
                     if(strOrdering.equalsIgnoreCase("causal"))
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~newMember~" + strVectorTS + "~" + strMessageData;
                     }
                     else
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~newMember~" + strMessageID + "~" + strMessageData;
                     }
                     byte[] msg = strMessage.getBytes();
                     // now use a DataOutputStream to print the byte array to the socket
                     DataOutputStream dosOut = new DataOutputStream(sockSendMessage.getOutputStream());
                     dosOut.write(msg);
                     dosOut.flush();
                     dosOut.close();
                     sockSendMessage.close();
                     /*
                     // next decorate the socket with a print writer
                     PrintWriter pwSendMessage = new PrintWriter(sockSendMessage.getOutputStream());
                     // next print out the message info to the socket
                     if(strOrdering.equalsIgnoreCase("causal")) {
                         pwSendMessage.println(strMemberName + ":" + intListenPort + "~" + strGroupNames + "~newMember~" + strVectorTS + "~" + strMessageData);
                     } else {
                         pwSendMessage.println(strMemberName + ":" + intListenPort + "~" + strGroupNames + "~newMember~" + strMessageID + "~" + strMessageData);
                     }
                     // pwSendMessage.println(strMessageData);
                     // force a flush of the print writer
                     pwSendMessage.flush();
                     // close and deallocate the socket and print writer
                     pwSendMessage.close();
                     sockSendMessage.close();
                      */
                  }
                  else
                  {
                     // here we are using UDP--get everything set up to transmit the message
                     String strMessage;
                     if(strOrdering.equalsIgnoreCase("causal"))
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~newMember~" + strVectorTS + "~" + strMessageData;
                     }
                     else
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~newMember~" + strMessageID + "~" + strMessageData;
                     }
                     byte[] msg = strMessage.getBytes();
                     // create a new datagram socket to send the message
                     DatagramSocket ds = new DatagramSocket();
                     InetAddress ia = InetAddress.getByName(strCurServer);
                     // create the outgoing data packet
                     DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,intCurListenPort);
                     // now send the packet to the server
                     ds.send(outPacket);
                     ds.close();
                  }
               }
            }
            else if(strMType.equalsIgnoreCase("sfnm"))
            {
               // here we need to send the sfnm (sender fail new member) message
               // next check to see that this is not the same server as this one (don't
               if(strCurServer.equalsIgnoreCase(this.strMemberName) && intCurListenPort == this.intListenPort)
               {
                  // do nothing here (it is the local one)
               }
               else if(strOrigSenderInfo.equalsIgnoreCase(strCurServer + ":" + strTemp))
               {
                  // this is the sender that has failed--do nothing here, either
               }
               else
               {
                  //System.out.println("Sending a sfnm message to " + strCurServer + ":" + intCurListenPort);
                  // add this member to the status vector
                  Vector vCurStatus = new Vector();
                  String strCurMember = strCurServer + ":" + intCurListenPort;
                  StringTokenizer stkDataID = new StringTokenizer(strMessageData,"|");
                  String strNewMemberInfo = stkDataID.nextToken();
                  String strJoinMessageID = stkDataID.nextToken();
                  vCurStatus.addElement(strCurMember);
                  vCurStatus.addElement("false");
                  vCurStatus.addElement(strNewMemberInfo);
                  vCurStatus.addElement(strJoinMessageID);
                  vTemp.addElement(vCurStatus);
                  // now we can send the message--determine if we are using TCP or UDP
                  if(strConType.equalsIgnoreCase("TCP"))
                  {
                     // here we can send the message--first open up a socket connection
                     Socket sockSendMessage = new Socket(strCurServer,intCurListenPort);
                     String strMessage;
                     if(strOrdering.equalsIgnoreCase("causal"))
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sfnm~" + strVectorTS + "~" + strOrigSenderInfo + "~" + strOrigMessageID + "~" + strMessageData;
                     }
                     else
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sfnm~" + strMessageID + "~" + strOrigSenderInfo + "~" + strOrigMessageID + "~" + strMessageData;
                     }
                     byte[] msg = strMessage.getBytes();
                     // now use a DataOutputStream to print the byte array to the socket
                     DataOutputStream dosOut = new DataOutputStream(sockSendMessage.getOutputStream());
                     dosOut.write(msg);
                     dosOut.flush();
                     dosOut.close();
                     sockSendMessage.close();
                     /*
                     // next decorate the socket with a print writer
                     PrintWriter pwSendMessage = new PrintWriter(sockSendMessage.getOutputStream());
                     // next print out the message info to the socket
                     if(strOrdering.equalsIgnoreCase("causal")) {
                         pwSendMessage.println(strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sfnm~" + strVectorTS + "~" + strOrigSenderInfo + "~" + strOrigMessageID + "~" + strMessageData);
                     } else {
                         pwSendMessage.println(strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sfnm~" + strMessageID + "~" + strOrigSenderInfo + "~" + strOrigMessageID + "~" + strMessageData);
                     }
                     // pwSendMessage.println(strMessageData);
                     // force a flush of the print writer
                     pwSendMessage.flush();
                     // close and deallocate the socket and print writer
                     pwSendMessage.close();
                     sockSendMessage.close();
                      */
                  }
                  else
                  {
                     // here we are using UDP--get everything set up to transmit the message
                     String strMessage;
                     if(strOrdering.equalsIgnoreCase("causal"))
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sfnm~" + strVectorTS + "~" + strOrigSenderInfo + "~" + strOrigMessageID + "~" + strMessageData;
                     }
                     else
                     {
                        strMessage = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~sfnm~" + strMessageID + "~" + strOrigSenderInfo + "~" + strOrigMessageID + "~" + strMessageData;
                     }
                     byte[] msg = strMessage.getBytes();
                     // create a new datagram socket to send the message
                     DatagramSocket ds = new DatagramSocket();
                     InetAddress ia = InetAddress.getByName(strCurServer);
                     // create the outgoing data packet
                     DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,intCurListenPort);
                     // now send the packet to the server
                     ds.send(outPacket);
                     ds.close();
                  }
               }
            }
         }
         //}

         this.intCurMessageID++;

         // next check to see if we are using total ordering
         if(strOrdering.equalsIgnoreCase("total"))
         {
            // check if we sent a "message" type
            if(strMType.equalsIgnoreCase("message"))
            {
               // here we first need to check to see if this member is the sequencer
               //  (the sequencer is the first member in the member group list)
               //synchronized (vGroup) {
               Vector vSequencerInfo = (Vector) vGroup.elementAt(0);
               // extract the information from the vector
               String strSequencerServer = (String) vSequencerInfo.elementAt(0);
               String strSequencerPort = (String) vSequencerInfo.elementAt(1);
               String strSequencerInfo = strSequencerServer + ":" + strSequencerPort;
               String strMemberInfo = strMemberName + ":" + Integer.toString(intListenPort);
               // check to see if this member is the sequencer
               if(strMemberInfo.equalsIgnoreCase(strSequencerInfo))
               {
                  // here this member is the sequencer--we need to send out the sequence message
                  // Thread.sleep(250);
                  //System.out.println("Sending sequence message for: " + strMemberInfo + "|" + strMessageID);
                  //intSequencer++;
                  sendMessage("","sequence",strMemberInfo,strMessageID);
               }
               //}
            }
         }

      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }

   /** Add a method that will start up the socket server
    */
   public void startListener()
   {
      try
      {
         // check to see if we need to create a TCP or UDP server
         if(strConType.equalsIgnoreCase("TCP"))
         {
            // check to see if the listener is running or not
            if(sockListener == null)
            {
               // here we need to start up the sockListener in its own thread
               sockListener = new SocketServer();
               Thread listenerThread = new Thread(sockListener);
               // now start up the thread (which will start the listener)
               listenerThread.start();
            }
         }
         else
         {
            // here the connection is UDP--check to see if the server is already running
            if(udpSockListener == null)
            {
               // here we need to start up the udpSockListener in its own thread
               udpSockListener = new UDPSocketServer();
               Thread listenerThread = new Thread(udpSockListener);
               // now start up the thread (which will start the listener)
               listenerThread.start();
            }
         }
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }

   /** Add a method that will stop the socket server
    */
   public void stopListener()
   {
      try
      {
         // check to see if we are using TCP or UDP
         if(strConType.equalsIgnoreCase("TCP"))
         {
            if(sockListener != null)
            {
               // call the stopRunning method
               sockListener.stopRunning();
               // now sleep for a little bit (to give it a chance to stop)
               Thread.sleep(500);
               // set the sockListener object to null
               sockListener = null;
            }
         }
         else
         {
            // here we are using UDP--stop the UDP server
            if(udpSockListener != null)
            {
               // call the stopRunning method
               udpSockListener.stopRunning();
               // now sleep for a little bit (to give it a chance to stop)
               Thread.sleep(500);
               // set the sockListener object to null
               udpSockListener = null;
            }
         }
      }
      catch (Exception e)
      {
         e.printStackTrace();
      }
   }

   /** A class that will monitor the responses from other group members
    */
   class newWatchMemberResponses
   {

      private Vector vStatusMonitor;
      private Vector vJoinMonitor;

      class MonitorStatus implements Runnable
      {

         public void run()
         {
            try
            {
               // here we need to set up a loop to scan the status buffer
               while(true)
               {
                  // here sleep for 1/3 the timeout
                  Thread.sleep(intTimeout / 3);
                  // next grab a copy of the status vector
                  Vector vSearchStatus = new Vector();
                  synchronized (vStatus)
                  {
                     vSearchStatus = (Vector) vStatus.clone();
                  }

                  // now we can connect an iterator to the cloned vector
                  //Iterator itVSS = vSearchStatus.iterator();
                  Enumeration itVSS = vSearchStatus.elements();
                  // set up a loop to check the contents to the vector
                  String strVSSIDs = ",";
                  while(itVSS.hasMoreElements())
                  {
                     //Vector vSTemp = (Vector) itVSS.next();
                     Vector vSTemp = (Vector) itVSS.nextElement();
                     // next we can extract the first element (which is the message id)
                     String strCurMessageID = (String) vSTemp.elementAt(0);
                     String strCurCausalMID = strCurMessageID;
                     // here we need to check and see if we need to bust up the timestamp
                     //  from the id
                     if(strOrdering.equalsIgnoreCase("causal"))
                     {
                        StringTokenizer stkMessageID = new StringTokenizer(strCurMessageID,"|");
                        // extract the message id from the timestamp
                        strCurMessageID = stkMessageID.nextToken();
                     }
                     strVSSIDs = strVSSIDs + strCurMessageID + ",";
                     // next we need to scan vStatusMonitor to find the matching record
                     // Iterator itVSM = vStatusMonitor.iterator();
                     Enumeration itVSM = vStatusMonitor.elements();
                     boolean blnFound = false;
                     while(itVSM.hasMoreElements())
                     {
                        Hashtable htCurElem = (Hashtable) itVSM.nextElement();
                        String strCheckMID = (String) htCurElem.get("messageid");
                        // check to see if we have found a match
                        if(strCurMessageID.equalsIgnoreCase(strCheckMID))
                        {
                           // here we found a match
                           blnFound = true;
                           String strNumPasses = (String) htCurElem.get("passes");
                           int intNumPasses = Integer.parseInt(strNumPasses);
                           // check to see if the message has been in the buffer
                           //  longer than 3 passes
                           if(intNumPasses == 3)
                           {
                              // here we can assume that the member has failed
                              //System.out.println("A member has not responded fast enough");
                              //System.exit(-1);

                              // next we need to extract the status info vector to determine who has failed
                              Vector vStatusInfo = (Vector) vSTemp.elementAt(1);

                              // here the status vector still exists, so at least one of the recepients has not
                              //  yet responded--determine who it is and store
                              String strRemoved = "";
                              for(int j=0;j<vStatusInfo.size();j++)
                              {
                                 // extract the current element
                                 Vector vSITemp = (Vector) vStatusInfo.elementAt(j);
                                 // extract the elements from this vector
                                 String strServerName = (String) vSITemp.elementAt(0);
                                 String strStatus = (String) vSITemp.elementAt(1);
                                 // check to see if the current group member has responded
                                 if(strStatus.equalsIgnoreCase("false"))
                                 {
                                    // this member has not responded in the required timeout
                                    //  period--assume this member has failed
                                    //System.out.println("Group member " + strServerName + " has failed and will be removed from the group.");
                                    // remove the member from the group
                                    removeMember(strServerName);
                                    // add the removed member to the removed string
                                    if(strRemoved.equalsIgnoreCase(""))
                                    {
                                       strRemoved = strServerName;
                                    }
                                    else
                                    {
                                       strRemoved = strRemoved + "," + strServerName;
                                    }
                                 }
                              }
                              // next we will need to send the garbage collection message to the new view (with the
                              //  removed members list
                              for(int j=0;j<vGroup.size();j++)
                              {
                                 // extract the current group member
                                 Vector vMember = (Vector) vGroup.elementAt(j);
                                 // extract the name and port of the current group member
                                 String strCurServer = (String) vMember.elementAt(0);
                                 String strCurPort = (String) vMember.elementAt(1);

                                 // check to see that this is not the local group member
                                 if(strCurServer.equalsIgnoreCase(strMemberName) && intListenPort == Integer.parseInt(strCurPort))
                                 {
                                    // local member--do nothing
                                 }
                                 else
                                 {
                                    // here we can send the garbage collection message
                                    // next we can build the header string for the gc message
                                    String strMessageHeader;
                                    if(strOrdering.equalsIgnoreCase("causal"))
                                    {
                                       strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                                          "gc~" + strCurCausalMID + "~" + strRemoved;
                                    }
                                    else
                                    {
                                       strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                                          "gc~" + strCurMessageID + "~" + strRemoved;
                                    }
                                    // check to see if we need to communicate via TCP or UDP
                                    if(strConType.equalsIgnoreCase("TCP"))
                                    {
                                       // next we need to open a socket connection to the sender
                                       Socket sock = new Socket(strCurServer,Integer.parseInt(strCurPort));
                                       byte[] msg = strMessageHeader.getBytes();
                                       // now use a DataOutputStream to print the byte array to the socket
                                       DataOutputStream dosOut = new DataOutputStream(sock.getOutputStream());
                                       dosOut.write(msg);
                                       dosOut.flush();
                                       dosOut.close();
                                       sock.close();

                                       /*// now decorate the socket with a print writer
                                       PrintWriter pout = new PrintWriter(sock.getOutputStream());
                                       // send the ack message back to the sender
                                       //System.out.println("Sending gc message to: " + strCurServer);
                                       pout.println(strMessageHeader);
                                       pout.flush();
                                       pout.close();
                                       sock.close();
                                        */
                                    }
                                    else
                                    {
                                       // here we need to use UDP to send the message
                                       byte[] msg = strMessageHeader.getBytes();
                                       // create a new datagram socket to send the message
                                       DatagramSocket ds = new DatagramSocket();
                                       InetAddress ia = InetAddress.getByName(strCurServer);
                                       // create the outgoing data packet
                                       DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,Integer.parseInt(strCurPort));
                                       // now send the packet to the server
                                       //System.out.println("Sending gc message to: " + strCurServer);
                                       ds.send(outPacket);
                                       ds.close();
                                    }
                                 }
                              }

                              synchronized (vStatus)
                              {
                                 vStatus.removeElement(vSTemp);
                              }
                              //itVSS.remove();
                              vSearchStatus.removeElement(vSTemp);

                           }
                           else
                           {
                              strNumPasses = Integer.toString(intNumPasses + 1);
                              htCurElem.put("passes",strNumPasses);
                           }
                           break;
                        }
                     }
                     // check to see if the entry was found
                     if(blnFound == false)
                     {
                        // here it wasn't found--create a new entry
                        Hashtable htNew = new Hashtable();
                        htNew.put("messageid",strCurMessageID);
                        htNew.put("passes","0");
                        // add the hashtable to the vector
                        vStatusMonitor.addElement(htNew);
                     }
                  }
                  // here need to insert code to clean out the old records in vStatusMonitor
                  //Iterator itVSM = vStatusMonitor.iterator();
                  Enumeration itVSM = vStatusMonitor.elements();
                  while(itVSM.hasMoreElements())
                  {
                     Hashtable htCurElm = (Hashtable) itVSM.nextElement();
                     String strCurID = (String) htCurElm.get("messageid");
                     if(strVSSIDs.indexOf("," + strCurID + ",") == -1)
                     {
                        // here we can remove this element (it no longer exists in vStatus)
                        //itVSM.remove();
                        vStatusMonitor.removeElement(htCurElm);
                     }
                  }
               }
            }
            catch (Exception e)
            {
               e.printStackTrace();
            }
         }

      }

      class MonitorJoin implements Runnable
      {

         public void run()
         {
            try
            {
               // here we need to set up a loop to scan the status buffer
               while(true)
               {
                  // here sleep for 1/3 the timeout
                  Thread.sleep(intTimeout / 3);
                  // next grab a copy of the status vector
                  Vector vSearchJoin = new Vector();
                  synchronized (vStatus)
                  {
                     vSearchJoin = (Vector) vJoinSent.clone();
                  }
                  // now we can connect an iterator to the cloned vector
                  //Iterator itVSJ = vSearchJoin.iterator();
                  Enumeration itVSJ = vSearchJoin.elements();
                  // set up a loop to check the contents to the vector
                  String strVSJIDs = ",";
                  while(itVSJ.hasMoreElements())
                  {
                     Vector vSTemp = (Vector) itVSJ.nextElement();
                     // next we can extract the first element (which is the message id)
                     String strCurMessageID = (String) vSTemp.elementAt(0);
                     String strCurCausalMID = strCurMessageID;
                     // here we need to check and see if we need to bust up the timestamp
                     //  from the id
                     if(strOrdering.equalsIgnoreCase("causal"))
                     {
                        StringTokenizer stkMessageID = new StringTokenizer(strCurMessageID,"|");
                        // extract the message id from the timestamp
                        strCurMessageID = stkMessageID.nextToken();
                     }
                     strVSJIDs = strVSJIDs + strCurMessageID + ",";
                     // next we need to scan vStatusMonitor to find the matching record
                     //Iterator itVJM = vJoinMonitor.iterator();
                     Enumeration itVJM = vJoinMonitor.elements();
                     boolean blnFound = false;
                     while(itVJM.hasMoreElements())
                     {
                        Hashtable htCurElem = (Hashtable) itVJM.nextElement();
                        String strCheckMID = (String) htCurElem.get("messageid");
                        // check to see if we have found a match
                        if(strCurMessageID.equalsIgnoreCase(strCheckMID))
                        {
                           // here we found a match
                           blnFound = true;
                           String strNumPasses = (String) htCurElem.get("passes");
                           int intNumPasses = Integer.parseInt(strNumPasses);
                           // check to see if the message has been in the buffer
                           //  longer than 3 passes
                           if(intNumPasses == 3)
                           {
                              // here we can assume that the member has failed
                              //System.out.println("A member has not responded fast enough");
                              //System.exit(-1);

                              // next we need to extract the status info vector to determine who has failed
                              Vector vStatusInfo = (Vector) vSTemp.elementAt(1);

                              // here the status vector still exists, so at least one of the recepients has not
                              //  yet responded--determine who it is and store
                              String strRemoved = "";
                              String strNewMemberInfo = "";
                              String strJoinMessageID = "";
                              for(int j=0;j<vStatusInfo.size();j++)
                              {
                                 // extract the current element
                                 Vector vSITemp = (Vector) vStatusInfo.elementAt(j);
                                 // extract the elements from this vector
                                 String strServerName = (String) vSITemp.elementAt(0);
                                 String strStatus = (String) vSITemp.elementAt(1);
                                 strNewMemberInfo = (String) vSITemp.elementAt(2);
                                 strJoinMessageID = (String) vSITemp.elementAt(3);
                                 // check to see if the current group member has responded
                                 if(strStatus.equalsIgnoreCase("false"))
                                 {
                                    // this member has not responded in the required timeout
                                    //  period--assume this member has failed
                                    //System.out.println("Group member " + strServerName + " has failed and will be removed from the group.");
                                    // remove the member from the group
                                    removeMember(strServerName);
                                    // add the removed member to the removed string
                                    if(strRemoved.equalsIgnoreCase(""))
                                    {
                                       strRemoved = strServerName;
                                    }
                                    else
                                    {
                                       strRemoved = strRemoved + "," + strServerName;
                                    }
                                 }

                              }

                              // next we will need to send the garbage collection message to the new view (with the
                              //  removed members list
                              for(int j=0;j<vGroup.size();j++)
                              {
                                 // extract the current group member
                                 Vector vMember = (Vector) vGroup.elementAt(j);
                                 // extract the name and port of the current group member
                                 String strCurServer = (String) vMember.elementAt(0);
                                 String strCurPort = (String) vMember.elementAt(1);

                                 if(strCurServer.equalsIgnoreCase(strMemberName) && intListenPort == Integer.parseInt(strCurPort))
                                 {
                                    // local member--do nothing
                                 }
                                 else if(strNewMemberInfo.equalsIgnoreCase(strCurServer + ":" + strCurPort))
                                 {
                                    // here we need to send the join success message to the new member
                                    //System.out.println("Sending join success message to new member: " + strNewMemberInfo);
                                    String strMessageHeader = null;
                                    if(strOrdering.equalsIgnoreCase("causal"))
                                    {
                                       strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                                          "jsuccess~" + strJoinMessageID + "~" + getVectorTimestamp();
                                    }
                                    else if(strOrdering.equalsIgnoreCase("total"))
                                    {
                                       // here we need to create a message header that includes the current sequence number
                                       //strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                       //                   "jsuccess~" + strJoinMessageID + "~" + strTotalSequence;
                                       strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                                          "jsuccess~" + strJoinMessageID + "~" + cTotalSequence.getCounter();
                                    }
                                    else
                                    {
                                       strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                                          "jsuccess~" + strJoinMessageID;
                                    }
                                    //String strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                    //                            "jsuccess~" + strJoinMessageID;
                                    // check to see if we need to use TCP or UDP to send the message
                                    if(strConType.equalsIgnoreCase("TCP"))
                                    {
                                       // next we need to open a socket connection to the sender
                                       Socket sock = new Socket(strCurServer,Integer.parseInt(strCurPort));
                                       byte[] msg = strMessageHeader.getBytes();
                                       // now use a DataOutputStream to print the byte array to the socket
                                       DataOutputStream dosOut = new DataOutputStream(sock.getOutputStream());
                                       dosOut.write(msg);
                                       dosOut.flush();
                                       dosOut.close();
                                       sock.close();
                                       /*
                                       // now decorate the socket with a print writer
                                       PrintWriter pout = new PrintWriter(sock.getOutputStream());
                                       // send the ack message back to the sender
                                       //System.out.println("Sending jsuccess message to: " + strCurServer);
                                       pout.println(strMessageHeader);
                                       pout.flush();
                                       pout.close();
                                       sock.close();
                                        */
                                    }
                                    else
                                    {
                                       // here we need to use UDP to send the message
                                       byte[] msg = strMessageHeader.getBytes();
                                       // create a new datagram socket to send the message
                                       DatagramSocket ds = new DatagramSocket();
                                       InetAddress ia = InetAddress.getByName(strCurServer);
                                       // create the outgoing data packet
                                       DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,Integer.parseInt(strCurPort));
                                       // now send the packet to the server
                                       //System.out.println("Sending jsuccess message to: " + strCurServer);
                                       ds.send(outPacket);
                                       ds.close();
                                    }
                                 }
                                 else
                                 {
                                    // here we can send the garbage collection message
                                    // next we can build the header string for the nmgc message
                                    String strMessageHeader;
                                    if(strOrdering.equalsIgnoreCase("causal"))
                                    {
                                       strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                                          "nmgc~" + strCurCausalMID + "~" + strRemoved;
                                    }
                                    else
                                    {
                                       strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                                          "nmgc~" + strCurMessageID + "~" + strRemoved;
                                    }
                                    // check to see if we need to use TCP or UDP to send the message
                                    if(strConType.equalsIgnoreCase("TCP"))
                                    {
                                       // next we need to open a socket connection to the sender
                                       Socket sock = new Socket(strCurServer,Integer.parseInt(strCurPort));
                                       byte[] msg = strMessageHeader.getBytes();
                                       // now use a DataOutputStream to print the byte array to the socket
                                       DataOutputStream dosOut = new DataOutputStream(sock.getOutputStream());
                                       dosOut.write(msg);
                                       dosOut.flush();
                                       dosOut.close();
                                       sock.close();
                                       /*
                                       // now decorate the socket with a print writer
                                       PrintWriter pout = new PrintWriter(sock.getOutputStream());
                                       // send the ack message back to the sender
                                       //System.out.println("Sending nmgc message to: " + strCurServer);
                                       pout.println(strMessageHeader);
                                       pout.flush();
                                       pout.close();
                                       sock.close();
                                        */
                                    }
                                    else
                                    {
                                       // here we need to use UDP to send the message
                                       byte[] msg = strMessageHeader.getBytes();
                                       // create a new datagram socket to send the message
                                       DatagramSocket ds = new DatagramSocket();
                                       InetAddress ia = InetAddress.getByName(strCurServer);
                                       // create the outgoing data packet
                                       DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,Integer.parseInt(strCurPort));
                                       // now send the packet to the server
                                       //System.out.println("Sending nmgc message to: " + strCurServer);
                                       ds.send(outPacket);
                                       ds.close();
                                    }
                                 }
                              }
                              synchronized (vJoinSent)
                              {
                                 vJoinSent.removeElement(vSTemp);
                              }

                           }
                           else
                           {
                              strNumPasses = Integer.toString(intNumPasses + 1);
                              htCurElem.put("passes",strNumPasses);
                           }
                           break;
                        }
                     }
                     // check to see if the entry was found
                     if(blnFound == false)
                     {
                        // here it wasn't found--create a new entry
                        Hashtable htNew = new Hashtable();
                        htNew.put("messageid",strCurMessageID);
                        htNew.put("passes","0");
                        // add the hashtable to the vector
                        vJoinMonitor.addElement(htNew);
                     }
                  }
                  // here need to insert code to clean out the old records in vStatusMonitor
                  // Iterator itVJM = vJoinMonitor.iterator();
                  Enumeration itVJM = vJoinMonitor.elements();
                  while(itVJM.hasMoreElements())
                  {
                     Hashtable htCurElm = (Hashtable) itVJM.nextElement();
                     String strCurID = (String) htCurElm.get("messageid");
                     if(strVSJIDs.indexOf("," + strCurID + ",") == -1)
                     {
                        // here we can remove this element (it no longer exists in vStatus)
                        // itVJM.remove();
                        vJoinMonitor.removeElement(htCurElm);
                     }
                  }
               }
            }
            catch (Exception e)
            {
               e.printStackTrace();
            }
         }

      }

      /* default constructor */
      public newWatchMemberResponses()
      {
         // initialize the monitor vectors
         vStatusMonitor = new Vector();
         vJoinMonitor = new Vector();
         // next start up two threads to monitor the buffers
         MonitorStatus ms = new MonitorStatus();
         Thread tMS = new Thread(ms);
         tMS.start();
         MonitorJoin mj = new MonitorJoin();
         Thread tMJ = new Thread(mj);
         tMJ.start();
      }

   }

   /** Add a class that will set up a UDP socket server on the listening port
    */
   class UDPSocketServer implements Runnable
   {

      private boolean blnTimeToQuit = false;

      public void stopRunning()
      {
         blnTimeToQuit = true;
      }

      public void run()
      {
         try
         {
            //System.out.println("Starting UDP Listening Server");
            // define some local variables
            DatagramSocket datagramSocket;
            DatagramPacket inDataPacket;
            // Incoming data buffer
            byte[] msg;
            // Allocate a socket to listen on the listening port for requests
            datagramSocket = new DatagramSocket(intListenPort);
            //System.out.println("UDP Server is active on port " + Integer.toString(intListenPort));
            // set up a loop to process the messages
            while(!blnTimeToQuit)
            {
               msg = new byte[Config.intPacketSize + 500];
               // receive the incoming data packet
               inDataPacket = new DatagramPacket(msg,msg.length);
               // get the message
               datagramSocket.receive(inDataPacket);
               // next we can extract the message from the packet and load it into the
               //  buffer
               String strTemp = new String(inDataPacket.getData(),0,inDataPacket.getLength());
               //synchronized(vNewMessages) {
               vNewMessages.addElement(strTemp);
               //}
            }
            //System.out.println("UDP Server is stopping");
         }
         catch (Exception e)
         {
            e.printStackTrace();
         }
      }
   }


   /** Add a class that will set up a socket server on the listening port
    */
   class SocketServer implements Runnable
   {

      private boolean blnTimeToQuit = false;

      public void stopRunning()
      {
         blnTimeToQuit = true;
      }

      public void run()
      {
         try
         {
            //System.out.println("Starting listening server");
            // create the server socket
            ServerSocket serverSock = new ServerSocket(intListenPort);
            // set up a loop to listen on the port forever
            while(!blnTimeToQuit)
            {
               // make a blocking call to wait for input on the socket
               Socket sock = serverSock.accept();
               // next we extract teh message and load it into the new message buffer
               DataInputStream disIn = new DataInputStream(sock.getInputStream());
               // next we read in a fixed packet size
               byte[] bInput = new byte[Config.intPacketSize + 500];
               int intNumBytes = disIn.read(bInput);
               byte[] bData = new byte[intNumBytes];
               System.arraycopy(bInput,0,bData,0,intNumBytes);
               // next we can convert the byte array into a string representation
               String strTemp = new String(bData);
               /*
               // next we extract the message and load it into the new message buffer
               BufferedReader brIn = new BufferedReader(new InputStreamReader(sock.getInputStream()));
               PrintWriter pwSOut = new PrintWriter(sock.getOutputStream());
               // read in a line from the socket
               String strTemp = brIn.readLine();
                */
               //synchronized(vNewMessages) {
               vNewMessages.addElement(strTemp);
               //}
               disIn.close();
               sock.close();
            }
            //System.out.println("Listening server is stopping");
         }
         catch (Exception e)
         {
            e.printStackTrace();
         }
      }

   }

   class newMonitorSender
   {

      private Vector vBufferMonitor;
      private Vector vJoinReceivedMonitor;

      public boolean CheckResponder(String strFailedMemberInfo)
      {
         // this method checks to see if the current member is the one who should
         //  respond if a sender failure is detected (the responder is the first one
         //  on the group list--just like the sequencer)
         Vector vGroupClone = new Vector();
         synchronized(vGroup)
         {
            vGroupClone = (Vector) vGroup.clone();
         }
         Vector vFirstInfo = (Vector) vGroupClone.elementAt(0);
         // extract the information from the vector
         String strFirstServer = (String) vFirstInfo.elementAt(0);
         String strFirstPort = (String) vFirstInfo.elementAt(1);
         String strFirstInfo = strFirstServer + ":" + strFirstPort;
         String strMemberInfo = strMemberName + ":" + Integer.toString(intListenPort);
         // check to see if this member is the sequencer
         //System.out.println("strMemberInfo is: " + strMemberInfo + ", strFirstInfo is: " + strFirstInfo);
         if(strMemberInfo.equalsIgnoreCase(strFirstInfo))
         {
            // here this member is the responder--return true
            return true;
         }
         else
         {
            // check to see if the responder is the one who has failed
            if(strFailedMemberInfo.equalsIgnoreCase(strFirstInfo))
            {
               // here we need to check the second member on the list
               Vector vSecondInfo = (Vector) vGroupClone.elementAt(1);
               // extract the information from the vector
               String strSecondServer = (String) vSecondInfo.elementAt(0);
               String strSecondPort = (String) vSecondInfo.elementAt(1);
               String strSecondInfo = strSecondServer + ":" + strSecondPort;
               if(strMemberInfo.equalsIgnoreCase(strSecondInfo))
               {
                  return true;
               }
               else
               {
                  return false;
               }
            }
            else
            {
               return false;
            }
         }
      }


      // implement a class to monitor the buffer
      class MonitorBuffer implements Runnable
      {

         public void run()
         {
            try
            {
               // here we need to set up a loop to run forever
               while(true)
               {
                  // here sleep for 1/3 the timeout
                  Thread.sleep(intTimeout / 3);
                  // next we need to grab a copy of the buffer
                  Vector vSearchBuffer = new Vector();
                  synchronized (vBuffer)
                  {
                     vSearchBuffer = (Vector) vBuffer.clone();
                  }
                  // now we can connect an iterator to the cloned vector
                  // Iterator itVSB = vSearchBuffer.iterator();
                  Enumeration itVSB = vSearchBuffer.elements();
                  // set up a loop to check the contents to the vector
                  String strVSBIDs = "~";
                  //System.out.println("vSearchBuffer size is: " + vSearchBuffer.size());
                  while(itVSB.hasMoreElements())
                  {
                     Vector vSTemp = (Vector) itVSB.nextElement();
                     // extract the sender info and message id
                     String strBufSender = (String) vSTemp.elementAt(0);
                     String strBufMessageID = (String) vSTemp.elementAt(1);
                     // next we need to store the sender and message id info
                     //  in a string (for garbage collection purposes later)
                     strVSBIDs = strVSBIDs + strBufSender + "," + strBufMessageID + "~";
                     // next we need to scan vBufferMonitor to find the matching record
                     // Iterator itVBM = vBufferMonitor.iterator();
                     Enumeration itVBM = vBufferMonitor.elements();
                     //System.out.println("vBufferMonitor size is: " + vBufferMonitor.size());
                     boolean blnFound = false;
                     while(itVBM.hasMoreElements())
                     {
                        Hashtable htCurElem = (Hashtable) itVBM.nextElement();
                        String strCheckMID = (String) htCurElem.get("messageid");
                        String strCheckSender = (String) htCurElem.get("sender");
                        // check to see if we have found a match
                        if(strBufMessageID.equalsIgnoreCase(strCheckMID) && strBufSender.equalsIgnoreCase(strCheckSender))
                        {
                           // here we found a match
                           blnFound = true;
                           String strNumPasses = (String) htCurElem.get("passes");
                           int intNumPasses = Integer.parseInt(strNumPasses);
                           //System.out.println("Number of passes is: " + intNumPasses);
                           // check to see if the message has been in the buffer
                           //  longer than 3 passes
                           if(intNumPasses == 3)
                           {
                              // here we can assume that the member has failed
                              //System.out.println("A sender has not responded fast enough");
                              //System.exit(-1);

                              // here we must assume that the sender has failed (since it did not respond in the required
                              //  timeout period).  Next we need to send out a senderfail message to the group and
                              //  then update our group view
                              // first extract the message data
                              if(CheckResponder(strBufSender) == true)
                              {
                                 String strMessageData = (String) vSTemp.elementAt(3);
                                 // now send the sf message to the remaining members
                                 sendMessage(strMessageData,"sf",strBufSender,strBufMessageID);
                                 // next we need to remove the failed sender from the group
                                 removeMember(strBufSender);
                              }
                           }
                           else
                           {
                              strNumPasses = Integer.toString(intNumPasses + 1);
                              htCurElem.put("passes",strNumPasses);
                           }
                        }
                     }

                     // check to see if the entry was found
                     if(blnFound == false)
                     {
                        // here it wasn't found--create a new entry
                        Hashtable htNew = new Hashtable();
                        htNew.put("messageid",strBufMessageID);
                        htNew.put("sender",strBufSender);
                        htNew.put("passes","0");
                        // add the hashtable to the vector
                        vBufferMonitor.addElement(htNew);
                     }
                  }
                  // here need to insert code to clean out the old records in vStatusMonitor
                  // Iterator itVBM = vBufferMonitor.iterator();
                  Enumeration itVBM = vBufferMonitor.elements();
                  //System.out.println("strVSBIDs is: " + strVSBIDs);
                  while(itVBM.hasMoreElements())
                  {
                     Hashtable htCurElm = (Hashtable) itVBM.nextElement();
                     String strCurID = (String) htCurElm.get("messageid");
                     String strCurServer = (String) htCurElm.get("sender");
                     //System.out.println("~" + strCurServer + "," + strCurID + "~");
                     if(strVSBIDs.indexOf("~" + strCurServer + "," + strCurID + "~") == -1)
                     {
                        // here we can remove this element (it no longer exists in vStatus)
                        // itVBM.remove();
                        vBufferMonitor.removeElement(htCurElm);
                     }
                  }
               }

            }
            catch (Exception e)
            {
               e.printStackTrace();
            }
         }

      }

      // implement a class to mointor the JoinReceived buffer
      class MonitorJoinReceived implements Runnable
      {

         public void run()
         {
            try
            {
               // here we need to set up a loop to run forever
               while(true)
               {
                  // here sleep for 1/3 the timeout
                  Thread.sleep(intTimeout / 3);
                  // next we need to grab a copy of the buffer
                  Vector vSearchJoinReceived = new Vector();
                  synchronized (vJoinReceived)
                  {
                     vSearchJoinReceived = (Vector) vJoinReceived.clone();
                  }
                  // now we can connect an iterator to the cloned vector
                  //Iterator itVSJ = vSearchJoinReceived.iterator();
                  Enumeration itVSJ = vSearchJoinReceived.elements();
                  // set up a loop to check the contents to the vector
                  String strVSJIDs = "~";
                  while(itVSJ.hasMoreElements())
                  {
                     Vector vSTemp = (Vector) itVSJ.nextElement();
                     // extract the sender info and message id
                     String strBufSender = (String) vSTemp.elementAt(0);
                     String strBufMessageID = (String) vSTemp.elementAt(1);
                     // next we need to store the sender and message id info
                     //  in a string (for garbage collection purposes later)
                     strVSJIDs = strVSJIDs + strBufSender + "," + strBufMessageID + "~";
                     // next we need to scan vBufferMonitor to find the matching record
                     // Iterator itVJM = vJoinReceivedMonitor.iterator();
                     Enumeration itVJM = vJoinReceivedMonitor.elements();
                     boolean blnFound = false;
                     while(itVJM.hasMoreElements())
                     {
                        Hashtable htCurElem = (Hashtable) itVJM.nextElement();
                        String strCheckMID = (String) htCurElem.get("messageid");
                        String strCheckSender = (String) htCurElem.get("sender");
                        // check to see if we have found a match
                        if(strBufMessageID.equalsIgnoreCase(strCheckMID) && strBufSender.equalsIgnoreCase(strCheckSender))
                        {
                           // here we found a match
                           blnFound = true;
                           String strNumPasses = (String) htCurElem.get("passes");
                           int intNumPasses = Integer.parseInt(strNumPasses);
                           // check to see if the message has been in the buffer
                           //  longer than 3 passes
                           if(intNumPasses == 3)
                           {
                              // here we can assume that the member has failed
                              //System.out.println("A sender has not responded fast enough");
                              //System.exit(-1);

                              // here we must assume that the sender has failed (since it did not respond in the required
                              //  timeout period).  Next we need to send out a senderfail message to the group and
                              //  then update our group view
                              // first extract the message data
                              if(CheckResponder(strBufSender) == true)
                              {
                                 // first extract the message data
                                 String strNewMemberInfo = (String) vSTemp.elementAt(3);
                                 String strJoinMessageID = (String) vSTemp.elementAt(4);
                                 // now send the sf message to the remaining members
                                 sendMessage(strNewMemberInfo + "|" + strJoinMessageID,"sfnm",strBufSender,strBufMessageID);
                                 // next we need to remove the failed sender from the group
                                 removeMember(strBufSender);
                              }
                           }
                           else
                           {
                              strNumPasses = Integer.toString(intNumPasses + 1);
                              htCurElem.put("passes",strNumPasses);
                           }
                        }
                     }

                     // check to see if the entry was found
                     if(blnFound == false)
                     {
                        // here it wasn't found--create a new entry
                        Hashtable htNew = new Hashtable();
                        htNew.put("messageid",strBufMessageID);
                        htNew.put("sender",strBufSender);
                        htNew.put("passes","0");
                        // add the hashtable to the vector
                        vJoinReceivedMonitor.addElement(htNew);
                     }
                  }
                  // here need to insert code to clean out the old records in vStatusMonitor
                  //Iterator itVJM = vJoinReceivedMonitor.iterator();
                  Enumeration itVJM = vJoinReceivedMonitor.elements();
                  while(itVJM.hasMoreElements())
                  {
                     Hashtable htCurElm = (Hashtable) itVJM.nextElement();
                     String strCurID = (String) htCurElm.get("messageid");
                     String strCurServer = (String) htCurElm.get("server");
                     if(strVSJIDs.indexOf("~" + strCurServer + "," + strCurID + "~") == -1)
                     {
                        // here we can remove this element (it no longer exists in vStatus)
                        // itVJM.remove();
                        vJoinReceivedMonitor.removeElement(htCurElm);
                     }
                  }
               }

            }
            catch (Exception e)
            {
               e.printStackTrace();
            }
         }

      }

      /* default constructor */
      public newMonitorSender()
      {
         // initialize the monitor vectors
         vBufferMonitor = new Vector();
         vJoinReceivedMonitor = new Vector();
         // next start up two threads to monitor the buffers
         MonitorBuffer mb = new MonitorBuffer();
         Thread tMB = new Thread(mb);
         tMB.start();
         MonitorJoinReceived mj = new MonitorJoinReceived();
         Thread tMJ = new Thread(mj);
         tMJ.start();
      }
   }

   /** add a class that will handle processing the messages that are
    *  received on the socket (either TCP or UDP)
    */
   class ProcessInput implements Runnable
   {

      public ProcessInput()
      {
         // nothing to initialize
      }

      public synchronized void run()
      {
         try
         {
            // set up a loop to run forever
            while(true)
            {
               // define a local variable to hold the data read in from the socket
               String strTemp = "";
               // get a clone of the new message vector to search
               Vector vNMClone = new Vector();
               synchronized(vNewMessages)
               {
                  vNMClone = (Vector) vNewMessages.clone();
               }
               // next set up a loop to process the messages in the cloned vector
               //Iterator itVNM = vNMClone.iterator();
               Enumeration itVNM = vNMClone.elements();
               boolean blnProcessed = false;
               while(itVNM.hasMoreElements())
               {
                  strTemp = (String) itVNM.nextElement();
                  String strRawMessage = strTemp;
                  // next we can determine what type of message this is
                  //  the format of the first line will be as follows:
                  //  sender~group list~type~message id
                  //  -- The sender will be the server name (ip address) of the sender
                  //  -- The group list will be a string that represents the group
                  //      (server name1:port1,server name2:port2, etc.)
                  //  -- The type will either be message, ack, gc (garbage collect), or sf (senderfail)
                  //  -- The message id will be an integer that is generated by the sender
                  //      if the message is of type message.  If the type is ack or
                  //      garbage collect then the message id will represent the id
                  //      of the message in the buffer that it represents
                  //  on the sf type message two extra parameters will be included--the original sender and original sender id
                  // use a string tokenizer to bust up the elements of the message
                  StringTokenizer stkMessageInfo = new StringTokenizer(strTemp,"~");
                  String strSender = stkMessageInfo.nextToken();
                  String strGroupList = stkMessageInfo.nextToken();
                  String strType = stkMessageInfo.nextToken();
                  String strMessageID = stkMessageInfo.nextToken();

                  // now we have extracted all the header information--next we need to determine
                  //  what we need to do based on the message type
                  boolean blnDeliverMessage = true;

                  if(strType.equalsIgnoreCase("sfnm"))
                  {
                     // here we have been sent a sender fail new member message--we first need to check if this
                     //  message has ever been received (by searching the history associated with
                     //  the failed sender
                     // first read in the extra header information
                     String strOrigSenderInfo = stkMessageInfo.nextToken();
                     String strOrigMessageID = stkMessageInfo.nextToken();
                     // first let's check the buffer to see if this message has already been
                     //  received and the member is waiting for the nmgc message from the sender
                     synchronized (vJoinReceived)
                     {
                        for(int i=0;i<vJoinReceived.size();i++)
                        {
                           // extract the current element from the vector
                           Vector vCurBufElem = (Vector) vJoinReceived.elementAt(i);
                           // now extract the message id and sender
                           String strCurBufSender = (String) vCurBufElem.elementAt(0);
                           String strCurBufMessageID = (String) vCurBufElem.elementAt(1);
                           // check to see if we have found a match
                           if(strOrigSenderInfo.equalsIgnoreCase(strCurBufSender) &&
                                 strOrigMessageID.equalsIgnoreCase(strCurBufMessageID))
                           {
                              // here we have found the matching message in the buffer--
                              //  we can now remove it
                              //synchronized (vJoinReceived) {
                              vJoinReceived.removeElement(vCurBufElem);
                              //}
                              // we also can break out of the for loop
                              break;
                           }
                        }
                     }
                     // now we can process the message like a new member message (since
                     //  if the member has already been added it will not be added again)
                     strType = "newMember";
                  }

                  if(strType.equalsIgnoreCase("join"))
                  {
                     blnProcessed = true;
                     // here we have received a join message from a non-member
                     //System.out.println("Received a join message from " + strSender);
                     // check to see if there are any other members to send the message to
                     if(vGroup.size() > 1)
                     {
                        // next we need to send the newMember message to all other members
                        sendMessage(strSender + "|" + strMessageID,"newMember","","");
                        // add the new group member (make sure it is not already a member)
                        if(findMember(strSender) == false)
                        {
                           // add the group member
                           String strNewGroup = strGroupNames + "," + strSender;
                           // setGroup(strNewGroup);
                           addGroupMember(strSender,strMessageID);
                        }
                     }
                     else
                     {
                        // here we just need to add the member and send a response back
                        if(findMember(strSender) == false)
                        {
                           // add the group member
                           String strNewGroup = strGroupNames + "," + strSender;
                           // setGroup(strNewGroup);
                           addGroupMember(strSender,strMessageID);
                        }
                        // here we need to send the join success message to the new member
                        //System.out.println("Sending join success message to new member: " + strSender);
                        String strMessageHeader;
                        if(strOrdering.equalsIgnoreCase("causal"))
                        {
                           strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                              "jsuccess~" + strMessageID + "~" + getVectorTimestamp();
                        }
                        else if(strOrdering.equalsIgnoreCase("total"))
                        {
                           // here we need to create a message header that includes the current sequence number
                           //strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                           //                   "jsuccess~" + strMessageID + "~" + strTotalSequence;
                           strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                              "jsuccess~" + strMessageID + "~" + cTotalSequence.getCounter();
                        }
                        else
                        {
                           strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                              "jsuccess~" + strMessageID;
                        }
                        StringTokenizer stkSB = new StringTokenizer(strSender,":");
                        String strCurServer = stkSB.nextToken();
                        String strCurPort = stkSB.nextToken();
                        // check to see if we need to use TCP or UDP to send the message
                        if(strConType.equalsIgnoreCase("TCP"))
                        {
                           // next we need to open a socket connection to the sender
                           Socket sock = new Socket(strCurServer,Integer.parseInt(strCurPort));
                           byte[] msg = strMessageHeader.getBytes();
                           // now use a DataOutputStream to print the byte array to the socket
                           DataOutputStream dosOut = new DataOutputStream(sock.getOutputStream());
                           dosOut.write(msg);
                           dosOut.flush();
                           dosOut.close();
                           sock.close();
                           /*
                           // now decorate the socket with a print writer
                           PrintWriter pout = new PrintWriter(sock.getOutputStream());
                           // send the ack message back to the sender
                           //System.out.println("Sending jsuccess message to: " + strCurServer);
                           pout.println(strMessageHeader);
                           pout.flush();
                           pout.close();
                           sock.close();
                            */
                        }
                        else
                        {
                           // here we need to use UDP to send the message
                           byte[] msg = strMessageHeader.getBytes();
                           // create a new datagram socket to send the message
                           DatagramSocket ds = new DatagramSocket();
                           InetAddress ia = InetAddress.getByName(strCurServer);
                           // create the outgoing data packet
                           DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,Integer.parseInt(strCurPort));
                           // now send the packet to the server
                           //System.out.println("Sending jsuccess message to: " + strCurServer);
                           ds.send(outPacket);
                           ds.close();
                        }
                     }
                  }
                  else if(strType.equalsIgnoreCase("newMember"))
                  {
                     // here we have recevied a new member message
                     String strDataID = stkMessageInfo.nextToken();
                     StringTokenizer stkDataID = new StringTokenizer(strDataID,"|");
                     String strNewMemberInfo = stkDataID.nextToken();
                     String strJoinMessageID = stkDataID.nextToken();
                     //System.out.println("Received a new member message from " + strSender);
                     //System.out.println("The new member info is: " + strNewMemberInfo);

                     // next we need to load this information into the vJoinReceived buffer
                     Vector vTemp = new Vector();
                     // add the sender
                     vTemp.addElement(strSender);
                     // add the message id
                     vTemp.addElement(strMessageID);
                     // add the group list (at this point we are ignoring this)
                     vTemp.addElement(strGroupList);
                     // add the new member connect info
                     vTemp.addElement(strNewMemberInfo);
                     // add the new member message id
                     vTemp.addElement(strJoinMessageID);
                     // now add this vector into the buffer
                     synchronized (vJoinReceived)
                     {
                        vJoinReceived.addElement(vTemp);
                     }

                     // next we need to add the new group member and send an ack back
                     // first check to see if this group member already exists
                     if(findMember(strNewMemberInfo) == false)
                     {
                        // add the group member
                        String strNewGroup = strGroupNames + "," + strNewMemberInfo;
                        // setGroup(strNewGroup);
                        addGroupMember(strNewMemberInfo,strJoinMessageID);
                     }

                     // next we may need to update the current message id for ordering
                     if(strOrdering.equalsIgnoreCase("fifo"))
                     {
                        // here we need to find the corresponding record in vFIFOMessageIDs
                        Vector vCurElement = null;
                        int k;
                        boolean blnFound = false;
                        synchronized (vFIFOMessageIDs)
                        {
                           //synchronized (vFIFOBuffer) {
                           for(k=0;k<vFIFOMessageIDs.size();k++)
                           {
                              // here we need to extract the current vector element
                              vCurElement = (Vector) vFIFOMessageIDs.elementAt(k);
                              // extract the sender name
                              String strCurSender = (String) vCurElement.elementAt(0);
                              // check to see if we have found a match
                              if(strCurSender.equalsIgnoreCase(strSender))
                              {
                                 // here we have found a match--indicate found and break out of the loop
                                 blnFound = true;
                                 break;
                              }
                           }
                           // check to see if the sender was found
                           if(blnFound == true)
                           {
                              // here check to see if this is the next message in the sequence
                              String strCurMessageID = (String) vCurElement.elementAt(1);
                              int intCurMessageID = Integer.parseInt(strCurMessageID);
                              int intMessageID = Integer.parseInt(strMessageID);
                              if((intMessageID - intCurMessageID) == 1)
                              {
                                 // here this is the next in the sequence so we need to update
                                 //  vFIFOMessageIDs
                                 // next we need to modify the vFIFOMessageIDs vector element
                                 Vector vTempA = new Vector();
                                 vTempA.addElement(strSender);
                                 vTempA.addElement(strMessageID);
                                 // here we can use a synchronized block on the vFIFOMessageIDs vector
                                 //synchronized (vFIFOMessageIDs) {
                                 vFIFOMessageIDs.setElementAt(vTempA,vFIFOMessageIDs.indexOf(vCurElement));
                                 //}
                                 // next we need to scan the vFIFOBuffer vector to see if there
                                 //  where any messages that where waiting for this message to arrive
                                 //   and be delivered so that they could be delivered
                                 boolean blnChanged = true;
                                 while(blnChanged == true)
                                 {
                                    blnChanged = false;
                                    // set up a loop to go through the vFIFOBuffer elements
                                    for(int j=0;j<vFIFOBuffer.size();j++)
                                    {
                                       Vector vCurFBElement = (Vector) vFIFOBuffer.elementAt(j);
                                       // extract the Sender
                                       String strCurSender = (String) vCurFBElement.elementAt(0);
                                       // check to see if we have found a match
                                       if(strCurSender.equalsIgnoreCase(strSender))
                                       {
                                          // we have found another message from this sender--extract the
                                          //  message id
                                          String strCurFMessageID = (String) vCurFBElement.elementAt(1);
                                          // check to see if this is the next message to be delivered
                                          int intCurFMessageID = Integer.parseInt(strCurFMessageID);
                                          if((intCurFMessageID - intMessageID) == 1)
                                          {
                                             // here this message needs to be delivered
                                             MessageEvent mCEvt = (MessageEvent) vCurFBElement.elementAt(2);
                                             String strDeliver = (String) vCurFBElement.elementAt(3);
                                             if(strDeliver.equalsIgnoreCase("yes"))
                                             {
                                                // make a copy of the listener list in case anyone adds/removes listeners
                                                Vector targets;
                                                synchronized (this)
                                                {
                                                   targets = (Vector) messageListeners.clone();
                                                }

                                                // walk through the listener list and call the receiveMessage method in each
                                                Enumeration e = targets.elements();
                                                while(e.hasMoreElements())
                                                {
                                                   MessageListener l = (MessageListener) e.nextElement();
                                                   l.receiveMessage(mCEvt);
                                                }
                                             }
                                             // next we need to modify the vFIFOMessageIDs vector element
                                             Vector vTempB = new Vector();
                                             vTempB.addElement(strSender);
                                             vTempB.addElement(strCurMessageID);
                                             // here we can use a synchronized block on the vFIFOMessageIDs vector
                                             //synchronized (vFIFOMessageIDs) {
                                             vFIFOMessageIDs.setElementAt(vTempB,vFIFOMessageIDs.indexOf(vCurElement));
                                             //}
                                             // next indicate that a change took place (so a rescan will take place)
                                             blnChanged = true;
                                             // reset the value of intMessageID
                                             intMessageID = intCurFMessageID;
                                             // remove the element from the buffer
                                             //synchronized (vFIFOBuffer) {
                                             vFIFOBuffer.removeElement(vCurFBElement);
                                             //}
                                             // next we need to subtract one from the loop variable (to
                                             //   make sure all elements get searched
                                             j--;
                                          }
                                       }
                                    }
                                 }
                              }
                              else
                              {
                                 // here we need to add this message to vFIFOBuffer as a message
                                 //  that does not need to be delivered (we are not worrying
                                 //  the ordering on join messages)
                                 Vector vTempA = new Vector();
                                 // add the sender, message id, and mEvt to the vector
                                 vTempA.addElement(strSender);
                                 vTempA.addElement(strMessageID);
                                 MessageEvent bogusME = new MessageEvent(null,strSender,strGroupList,strMessageID,"bogus");
                                 vTempA.addElement(bogusME);
                                 vTempA.addElement("no");
                                 vFIFOBuffer.addElement(vTempA);
                              }
                           }
                           //}
                        }
                     }
                     else if(strOrdering.equalsIgnoreCase("causal"))
                     {
                        synchronized (vCausalMessageIDs)
                        {
                           //synchronized (vCausalBuffer) {
                           // here we need to check and see if we need to update vCausalMessageIDs
                           //  with the latest message id, or buffer the message

                           // next we need to bust up the message ID field to extract the vector timestamp
                           //  elements
                           StringTokenizer stkMessageID = new StringTokenizer(strMessageID,"|,");
                           // the first element will contain the message id from the sender (we can
                           //  ignore this one for now)
                           stkMessageID.nextToken();
                           // next we can set up a loop to extract the server names and their
                           //  corresponding sequence numbers and determine if it is okay to deliver
                           //  this message
                           boolean blnOkayToDeliver = true;
                           // define a vector to hold the potential update
                           Vector vVUpdate = new Vector();
                           // define a vector to hold the element that will possibly get updated
                           Vector vToBeUpdated = new Vector();
                           while(stkMessageID.hasMoreTokens())
                           {
                              // extract the server info
                              String strCurServerInfo = stkMessageID.nextToken();
                              // extract the sequence number for this member
                              String strCurSequenceNum = stkMessageID.nextToken();
                              //System.out.println("Processing the sequence number: " + strCurSequenceNum + " for Member: " + strCurServerInfo);
                              // next we need to find the corresponding element in vCausalMessageIDs
                              for(int i=0;i<vCausalMessageIDs.size();i++)
                              {
                                 // first extract the vector element
                                 Vector vCurElem = (Vector) vCausalMessageIDs.elementAt(i);
                                 // next extract the member info
                                 String strCurMemberInfo = (String) vCurElem.elementAt(0);
                                 // next extact the current sequence number for this member
                                 String strCurLastRecSeq = (String) vCurElem.elementAt(1);
                                 if(strCurMemberInfo.equalsIgnoreCase(strCurServerInfo))
                                 {
                                    // here we have found a match--next we need to check the sequence
                                    //  numbers to see if the message can be delivered.  Here if the
                                    //  the current member is the sender of the message then the
                                    //  sequence number will be one higher than what we have in the
                                    //  buffer.  If the current member is not the sender of the message
                                    //  then the sequence number should be the same as what we have in
                                    //  the buffer
                                    //System.out.println("Info in Causal vector is: " + strCurMemberInfo + " - " + strCurLastRecSeq);
                                    if(strCurMemberInfo.equalsIgnoreCase(strSender))
                                    {
                                       // check to see that the sequence number is one larger than what is
                                       //  in the buffer
                                       int intCurSequenceNum = Integer.parseInt(strCurSequenceNum);
                                       int intCurLastRecSeq = Integer.parseInt(strCurLastRecSeq);
                                       // first check to see if we have ever received any messages from this user
                                       if(intCurLastRecSeq == -1)
                                       {
                                          // here this is the first message we have received from this user (assume it is okay to
                                          //  deliver)
                                          //System.out.println("Okay to deliver for " + strCurMemberInfo);
                                          // next we need to update the last received message id in the causal vector
                                          vVUpdate = new Vector();
                                          vVUpdate.addElement(strCurMemberInfo);
                                          vVUpdate.addElement(strCurSequenceNum);
                                          // next store the element to be updated
                                          vToBeUpdated = vCurElem;
                                       }
                                       else
                                       {
                                          if((intCurSequenceNum - intCurLastRecSeq) == 1)
                                          {
                                             //System.out.println("Okay to deliver for " + strCurMemberInfo);
                                             // next we need to update the last received message id in the causal vector
                                             vVUpdate = new Vector();
                                             vVUpdate.addElement(strCurMemberInfo);
                                             vVUpdate.addElement(strCurSequenceNum);
                                             // next store the element to be updated
                                             vToBeUpdated = vCurElem;
                                          }
                                          else
                                          {
                                             //System.out.println("Not okay to deliver for " + strCurMemberInfo);
                                             blnOkayToDeliver = false;
                                          }
                                       }
                                    }
                                    else
                                    {
                                       // check to see that the sequence number matches the one
                                       //  in the buffer
                                       int intCurSequenceNum = Integer.parseInt(strCurSequenceNum);
                                       int intCurLastRecSeq = Integer.parseInt(strCurLastRecSeq);
                                       if((intCurSequenceNum - intCurLastRecSeq) == 0)
                                       {
                                          //System.out.println("Okay to deliver for " + strCurMemberInfo);
                                       }
                                       else
                                       {
                                          //System.out.println("Not okay to deliver for " + strCurMemberInfo);
                                          blnOkayToDeliver = false;
                                       }
                                    }
                                 }
                              }
                           }

                           // check to see if we need to deliver this message or buffer it
                           if(blnOkayToDeliver == true)
                           {
                              // now replace the element in vCausalMessageIDs
                              //synchronized (vCausalMessageIDs) {
                              vCausalMessageIDs.setElementAt(vVUpdate,vCausalMessageIDs.indexOf(vToBeUpdated));
                              //}

                              // next we need to scan the vCausalBuffer vector to see if there
                              //  where any messages that where waiting for this message to arrive
                              //   and be delivered so that they could be delivered

                              // next we need to set up a loop to see if any other messages in the buffer
                              //  can now be delivered
                              boolean blnChanged = true;
                              while(blnChanged == true)
                              {
                                 blnChanged = false;
                                 // set up a loop to go through the buffer looking for a message that is
                                 //  now ready to deliver
                                 for(int j=0;j<vCausalBuffer.size();j++)
                                 {
                                    Vector vCheckEventInfo = (Vector) vCausalBuffer.elementAt(j);
                                    // next extract the Sender and Message ID
                                    String strCheckSender = (String) vCheckEventInfo.elementAt(0);
                                    String strCheckMessageID = (String) vCheckEventInfo.elementAt(1);
                                    MessageEvent mCheckEvt = (MessageEvent) vCheckEventInfo.elementAt(2);
                                    String strCheckDeliver = (String) vCheckEventInfo.elementAt(3);
                                    // here the message ID will be a delimited string--the first element
                                    //  will be the sequence number from the sender and the rest of
                                    //  the elements will be the vector timestamp that contains the
                                    //  causal history for this message
                                    // next we need to bust up the message ID field to extract the vector timestamp
                                    //  elements
                                    StringTokenizer stkCheckMessageID = new StringTokenizer(strCheckMessageID,"|,");
                                    // the first element will contain the message id from the sender (we can
                                    //  ignore this one for now)
                                    stkCheckMessageID.nextToken();
                                    // next we can set up a loop to extract the server names and their
                                    //  corresponding sequence numbers and determine if it is okay to deliver
                                    //  this message
                                    boolean blnCheckOkayToDeliver = true;
                                    // define a vector to hold the potential update
                                    Vector vCheckVUpdate = new Vector();
                                    // define a vector to hold the element that will possibly get updated
                                    Vector vCheckToBeUpdated = new Vector();
                                    while(stkCheckMessageID.hasMoreTokens())
                                    {
                                       // extract the server info
                                       String strCheckCurServerInfo = stkCheckMessageID.nextToken();
                                       // extract the sequence number for this member
                                       String strCheckCurSequenceNum = stkCheckMessageID.nextToken();
                                       //System.out.println("Processing the sequence number: " + strCheckCurSequenceNum + " for Member: " + strCheckCurServerInfo);
                                       // next we need to find the corresponding element in vCausalMessageIDs
                                       for(int i=0;i<vCausalMessageIDs.size();i++)
                                       {
                                          // first extract the vector element
                                          Vector vCheckCurElem = (Vector) vCausalMessageIDs.elementAt(i);
                                          // next extract the member info
                                          String strCheckCurMemberInfo = (String) vCheckCurElem.elementAt(0);
                                          // next extact the current sequence number for this member
                                          String strCheckCurLastRecSeq = (String) vCheckCurElem.elementAt(1);
                                          if(strCheckCurMemberInfo.equalsIgnoreCase(strCheckCurServerInfo))
                                          {
                                             // here we have found a match--next we need to check the sequence
                                             //  numbers to see if the message can be delivered.  Here if the
                                             //  the current member is the sender of the message then the
                                             //  sequence number will be one higher than what we have in the
                                             //  buffer.  If the current member is not the sender of the message
                                             //  then the sequence number should be the same as what we have in
                                             //  the buffer
                                             //System.out.println("Info in Causal vector is: " + strCheckCurMemberInfo + " - " + strCheckCurLastRecSeq);
                                             if(strCheckCurMemberInfo.equalsIgnoreCase(strCheckSender))
                                             {
                                                // check to see that the sequence number is one larger than what is
                                                //  in the buffer
                                                int intCheckCurSequenceNum = Integer.parseInt(strCheckCurSequenceNum);
                                                int intCheckCurLastRecSeq = Integer.parseInt(strCheckCurLastRecSeq);
                                                // first check to see if we have ever received any messages from this user
                                                if(intCheckCurLastRecSeq == -1)
                                                {
                                                   // here this is the first message we have received from this user (assume it is okay to
                                                   //  deliver)
                                                   //System.out.println("Okay to deliver for " + strCheckCurMemberInfo);
                                                   // next we need to update the last received message id in the causal vector
                                                   vCheckVUpdate = new Vector();
                                                   vCheckVUpdate.addElement(strCheckCurMemberInfo);
                                                   vCheckVUpdate.addElement(strCheckCurSequenceNum);
                                                   // next store the element to be updated
                                                   vCheckToBeUpdated = vCheckCurElem;
                                                }
                                                else
                                                {
                                                   if((intCheckCurSequenceNum - intCheckCurLastRecSeq) == 1)
                                                   {
                                                      //System.out.println("Okay to deliver for " + strCheckCurMemberInfo);
                                                      // next we need to update the last received message id in the causal vector
                                                      vCheckVUpdate = new Vector();
                                                      vCheckVUpdate.addElement(strCheckCurMemberInfo);
                                                      vCheckVUpdate.addElement(strCheckCurSequenceNum);
                                                      // next store the element to be updated
                                                      vCheckToBeUpdated = vCheckCurElem;
                                                   }
                                                   else
                                                   {
                                                      //System.out.println("Not okay to deliver for " + strCheckCurMemberInfo);
                                                      blnCheckOkayToDeliver = false;
                                                   }
                                                }
                                             }
                                             else
                                             {
                                                // check to see that the sequence number matches the one
                                                //  in the buffer
                                                int intCheckCurSequenceNum = Integer.parseInt(strCheckCurSequenceNum);
                                                int intCheckCurLastRecSeq = Integer.parseInt(strCheckCurLastRecSeq);
                                                if((intCheckCurSequenceNum - intCheckCurLastRecSeq) == 0)
                                                {
                                                   //System.out.println("Okay to deliver for " + strCheckCurMemberInfo);
                                                }
                                                else
                                                {
                                                   //System.out.println("Not okay to deliver for " + strCheckCurMemberInfo);
                                                   blnCheckOkayToDeliver = false;
                                                }
                                             }
                                          }
                                       }
                                    }

                                    // check to see if we need to deliver this message or buffer it
                                    if(blnCheckOkayToDeliver == true)
                                    {
                                       // it is okay to deliver this message to all listeners
                                       blnChanged = true;
                                       if(strCheckDeliver.equalsIgnoreCase("yes"))
                                       {
                                          Vector vChecktargets;
                                          synchronized (this)
                                          {
                                             vChecktargets = (Vector) messageListeners.clone();
                                          }

                                          // walk through the listener list and call the receiveMessage method in each
                                          Enumeration eC = vChecktargets.elements();
                                          while(eC.hasMoreElements())
                                          {
                                             MessageListener l = (MessageListener) eC.nextElement();
                                             l.receiveMessage(mCheckEvt);
                                          }
                                       }

                                       // next we need to update vCausalMessageIDs
                                       // now replace the element in vCausalMessageIDs
                                       //synchronized (vCausalMessageIDs) {
                                       vCausalMessageIDs.setElementAt(vCheckVUpdate,vCausalMessageIDs.indexOf(vCheckToBeUpdated));
                                       //}
                                    }
                                 }
                              }

                           }
                           else
                           {
                              // here we need to store the message in the buffer (it cannot be delivered yet)
                              Vector vTempA = new Vector();
                              // add the sender, message id, and mEvt to the vector
                              vTempA.addElement(strSender);
                              vTempA.addElement(strMessageID);
                              MessageEvent bogusME = new MessageEvent(null,strSender,strGroupList,strMessageID,"bogus");
                              vTempA.addElement(bogusME);
                              vTempA.addElement("no");
                              vCausalBuffer.addElement(vTempA);
                           }
                           //}
                        }
                     }

                     // now send back an nmack message
                     StringTokenizer stkSender = new StringTokenizer(strSender,":");
                     String strServerName = stkSender.nextToken();
                     int intDestPort = Integer.parseInt(stkSender.nextToken());
                     // next we can build the header string for the ack message
                     String strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupList + "~" +
                                               "nmack~" + strMessageID;
                     // check to see if we need to use TCP or UDP to send the message
                     if(strConType.equalsIgnoreCase("TCP"))
                     {
                        // next we need to open a socket connection to the sender
                        Socket sock = new Socket(strServerName,intDestPort);
                        byte[] msg = strMessageHeader.getBytes();
                        // now use a DataOutputStream to print the byte array to the socket
                        DataOutputStream dosOut = new DataOutputStream(sock.getOutputStream());
                        dosOut.write(msg);
                        dosOut.flush();
                        dosOut.close();
                        sock.close();
                        /*
                        // now decorate the socket with a print writer
                        PrintWriter pout = new PrintWriter(sock.getOutputStream());
                        // send the ack message back to the sender
                        //System.out.println("Sending nmack message back to: " + strSender);
                        pout.println(strMessageHeader);
                        pout.flush();
                        pout.close();
                        sock.close();
                         */
                     }
                     else
                     {
                        // here we need to use UDP to send the message
                        byte[] msg = strMessageHeader.getBytes();
                        // create a new datagram socket to send the message
                        DatagramSocket ds = new DatagramSocket();
                        InetAddress ia = InetAddress.getByName(strServerName);
                        // create the outgoing data packet
                        DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,intDestPort);
                        // now send the packet to the server
                        //System.out.println("Sending nmack message back to: " + strSender);
                        ds.send(outPacket);
                        ds.close();
                     }


                     // here we need to kick off a thread to handle if the sender times out
                     //   and never sends the nmgc message
                     /*
                     MonitorSenderNM ms = new MonitorSenderNM(strSender,strMessageID);
                     Thread tMS = new Thread(ms);
                     // start the thread
                     tMS.start();
                      */
                     blnProcessed = true;
                  }
                  else if(strType.equalsIgnoreCase("nmack"))
                  {
                     // this is an ack for a newMember message
                     //System.out.println("Received an nmack message from :" + strSender);
                     //System.out.println("The message id is: " + strMessageID);
                     // here we need to update the status buffer and see if all nmacks have
                     //  been received
                     int i;
                     boolean blnFound = false;
                     Vector vStatusInfo = new Vector();
                     Vector vSTemp = new Vector();
                     synchronized (vJoinSent)
                     {
                        for(i=0;i<vJoinSent.size();i++)
                        {
                           // first we need to pull the current vector element off of the list
                           vSTemp = (Vector) vJoinSent.elementAt(i);
                           // next we can extract the first element (which is the message id)
                           String strCurMessageID = (String) vSTemp.elementAt(0);
                           // check to see if we have found a match
                           if(strCurMessageID.equalsIgnoreCase(strMessageID))
                           {
                              // here we have found a match--extract the status vector
                              vStatusInfo = (Vector) vSTemp.elementAt(1);
                              // indicate that the element was found
                              blnFound = true;
                              // break out of the loop
                              break;
                           }
                        }

                        // check to see that the corresponding message was found
                        if(blnFound == true)
                        {
                           // create a string to hold the connect info for the new member
                           String strNewMemberInfo = "";
                           String strJoinMessageID = "";
                           // create a boolean variable to indicate whether all acks have been
                           //  received
                           boolean blnAllReceived = true;
                           // next we can update the element that corresponds to this sender
                           for(int j=0;j<vStatusInfo.size();j++)
                           {
                              // extract the current element
                              Vector vSITemp = (Vector) vStatusInfo.elementAt(j);
                              // extract the elements from this vector
                              String strServerName = (String) vSITemp.elementAt(0);
                              String strStatus = (String) vSITemp.elementAt(1);
                              strNewMemberInfo = (String) vSITemp.elementAt(2);
                              strJoinMessageID = (String) vSITemp.elementAt(3);
                              // check to see if the server name matches the sender
                              if(strServerName.equalsIgnoreCase(strSender))
                              {
                                 // here we need to update strStatus to true
                                 strStatus = "true";
                                 // update the status vector
                                 // first we need to make a clone of the vector that was found
                                 Vector vNewStatus = (Vector) vSITemp.clone();
                                 vNewStatus.setElementAt(strStatus,1);
                                 synchronized (vStatusInfo)
                                 {
                                    vStatusInfo.setElementAt(vNewStatus,vStatusInfo.indexOf(vSITemp));
                                 }
                              }
                              else
                              {
                                 // check to see if an ack has been received by this member
                                 if(strStatus.equalsIgnoreCase("false"))
                                 {
                                    // indicate that all acks have not been received
                                    blnAllReceived = false;
                                 }
                              }
                           }
                           // next we need to load the statusinfo vector back into status vector
                           Vector vSLoad = new Vector();
                           vSLoad.addElement(strMessageID);
                           vSLoad.addElement(vStatusInfo);
                           //synchronized (vJoinSent) {
                           vJoinSent.setElementAt(vSLoad,vJoinSent.indexOf(vSTemp));
                           //}
                           // next check to see if all acks have been received
                           if(blnAllReceived == true)
                           {
                              //System.out.println("All nmacks have been received for message: " + strMessageID);
                              // here we need to send the garbage collection message to all members
                              //synchronized (vGroup) {
                              for(int j=0;j<vGroup.size();j++)
                              {
                                 // extract the current group member
                                 Vector vMember = (Vector) vGroup.elementAt(j);
                                 // extract the name and port of the current group member
                                 String strCurServer = (String) vMember.elementAt(0);
                                 String strCurPort = (String) vMember.elementAt(1);
                                 // check to see that this is not the local group member
                                 if(strCurServer.equalsIgnoreCase(strMemberName) && intListenPort == Integer.parseInt(strCurPort))
                                 {
                                    // local member--do nothing
                                 }
                                 else if(strNewMemberInfo.equalsIgnoreCase(strCurServer + ":" + strCurPort))
                                 {
                                    // here we need to send the join success message to the new member
                                    //System.out.println("Sending join success message to new member: " + strNewMemberInfo);
                                    String strMessageHeader;
                                    if(strOrdering.equalsIgnoreCase("causal"))
                                    {
                                       strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                                          "jsuccess~" + strJoinMessageID + "~" + getVectorTimestamp();
                                    }
                                    else if(strOrdering.equalsIgnoreCase("total"))
                                    {
                                       // here we need to return the current total sequence number
                                       //strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                       //                   "jsuccess~" + strJoinMessageID + "~" + strTotalSequence;
                                       strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                                          "jsuccess~" + strJoinMessageID + "~" + cTotalSequence.getCounter();
                                    }
                                    else
                                    {
                                       strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupNames + "~" +
                                                          "jsuccess~" + strJoinMessageID;
                                    }
                                    // check to see if we need to use TCP or UDP to send the message
                                    if(strConType.equalsIgnoreCase("TCP"))
                                    {
                                       // next we need to open a socket connection to the sender
                                       Socket sock = new Socket(strCurServer,Integer.parseInt(strCurPort));
                                       byte[] msg = strMessageHeader.getBytes();
                                       // now use a DataOutputStream to print the byte array to the socket
                                       DataOutputStream dosOut = new DataOutputStream(sock.getOutputStream());
                                       dosOut.write(msg);
                                       dosOut.flush();
                                       dosOut.close();
                                       sock.close();
                                       /*
                                       // now decorate the socket with a print writer
                                       PrintWriter pout = new PrintWriter(sock.getOutputStream());
                                       // send the ack message back to the sender
                                       //System.out.println("Sending jsuccess message to: " + strCurServer);
                                       pout.println(strMessageHeader);
                                       pout.flush();
                                       pout.close();
                                       sock.close();
                                        */
                                    }
                                    else
                                    {
                                       // here we need to use UDP to send the message
                                       byte[] msg = strMessageHeader.getBytes();
                                       // create a new datagram socket to send the message
                                       DatagramSocket ds = new DatagramSocket();
                                       InetAddress ia = InetAddress.getByName(strCurServer);
                                       // create the outgoing data packet
                                       DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,Integer.parseInt(strCurPort));
                                       // now send the packet to the server
                                       //System.out.println("Sending jsuccess message to: " + strCurServer);
                                       ds.send(outPacket);
                                       ds.close();
                                    }
                                 }
                                 else
                                 {
                                    // here we can send the garbage collection message
                                    // next we can build the header string for the nmgc message
                                    String strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupList + "~" +
                                                              "nmgc~" + strMessageID;
                                    // check to see if we need to use TCP or UDP to send the message
                                    if(strConType.equalsIgnoreCase("TCP"))
                                    {
                                       // next we need to open a socket connection to the sender
                                       Socket sock = new Socket(strCurServer,Integer.parseInt(strCurPort));
                                       byte[] msg = strMessageHeader.getBytes();
                                       // now use a DataOutputStream to print the byte array to the socket
                                       DataOutputStream dosOut = new DataOutputStream(sock.getOutputStream());
                                       dosOut.write(msg);
                                       dosOut.flush();
                                       dosOut.close();
                                       sock.close();
                                       /*
                                       // now decorate the socket with a print writer
                                       PrintWriter pout = new PrintWriter(sock.getOutputStream());
                                       // send the ack message back to the sender
                                       //System.out.println("Sending nmgc message to: " + strCurServer);
                                       pout.println(strMessageHeader);
                                       pout.flush();
                                       pout.close();
                                       sock.close();
                                        */
                                    }
                                    else
                                    {
                                       // here we need to use UDP to send the message
                                       byte[] msg = strMessageHeader.getBytes();
                                       // create a new datagram socket to send the message
                                       DatagramSocket ds = new DatagramSocket();
                                       InetAddress ia = InetAddress.getByName(strCurServer);
                                       // create the outgoing data packet
                                       DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,Integer.parseInt(strCurPort));
                                       // now send the packet to the server
                                       //System.out.println("Sending nmgc message to: " + strCurServer);
                                       ds.send(outPacket);
                                       ds.close();
                                    }
                                 }
                              }
                              //}
                              // next we can remove the status vector entry
                              //synchronized (vJoinSent) {
                              vJoinSent.removeElement(vSLoad);
                              //}
                           }
                           blnProcessed = true;
                        }
                     }
                  }
                  else if(strType.equalsIgnoreCase("nmgc"))
                  {
                     //System.out.println("Received nmgc message from: " + strSender);
                     // here we need to remove the corresponding message from vJoinReceived
                     synchronized (vJoinReceived)
                     {
                        for(int i=0;i<vJoinReceived.size();i++)
                        {
                           Vector vCurElem = (Vector) vJoinReceived.elementAt(i);
                           // extract the sender and message id
                           String strCurElemSender = (String) vCurElem.elementAt(0);
                           String strCurMessageID = (String) vCurElem.elementAt(1);
                           if(strCurElemSender.equalsIgnoreCase(strSender) && strCurMessageID.equalsIgnoreCase(strMessageID))
                           {
                              // the element has been found--remove it and break out of the loop
                              //synchronized (vJoinReceived) {
                              vJoinReceived.removeElement(vCurElem);
                              //}
                              blnProcessed = true;
                              break;
                           }
                        }
                     }
                  }
                  else if(strType.equalsIgnoreCase("jsuccess"))
                  {
                     //System.out.println("Received a jsuccess message from: " + strSender);
                     //System.out.println("Message ID was: " + strMessageID);
                     //System.out.println("The new group is: " + strGroupList);
                     // install the group that was passed in
                     setGroup(strGroupList);
                     // set up a loop to find the corresponding record in vJoinSent
                     String strContactInfo = "";
                     synchronized (vJoinSent)
                     {
                        for(int i=0;i<vJoinSent.size();i++)
                        {
                           // extract the current vector element
                           Vector vJoinElement = (Vector) vJoinSent.elementAt(i);
                           // extract the message id from the join element
                           String strCurMessageID = (String) vJoinElement.elementAt(0);
                           // check to see if we have found a match
                           if(strCurMessageID.equalsIgnoreCase(strMessageID))
                           {
                              // here we need to extract the contact info
                              strContactInfo = (String) vJoinElement.elementAt(1);
                              // here we need to remove the element from the vector
                              //synchronized (vJoinSent) {
                              vJoinSent.removeElement(vJoinElement);
                              //}
                              // now we can break out of the loop
                              blnProcessed = true;
                              break;
                           }
                        }
                     }

                     if(strOrdering.equalsIgnoreCase("causal"))
                     {
                        synchronized (vCausalMessageIDs)
                        {
                           // extract the current vector timestamp
                           String strVTS = stkMessageInfo.nextToken();
                           //System.out.println("The current vector timestamp is: " + strVTS);
                           // next need to install the vector timestamp that was passed back
                           //vCausalMessageIDs.clear();
                           vCausalMessageIDs.removeAllElements();
                           StringTokenizer stkTimestamp = new StringTokenizer(strVTS,",");
                           while(stkTimestamp.hasMoreTokens())
                           {
                              String strCurDestInfo = stkTimestamp.nextToken();
                              // connect a string tokenizer to this string to get the member info
                              //  and id
                              StringTokenizer stkCurDest = new StringTokenizer(strCurDestInfo,"|");
                              // extract the server name and message id
                              String strCurDest = stkCurDest.nextToken();
                              String strCurID = stkCurDest.nextToken();
                              // here we need to check if this is the sender
                              //System.out.println(strCurDest + "," + strContactInfo);
                              if(strCurDest.equals(strContactInfo))
                              {
                                 // here we need to decrease the value of strCurID by one
                                 int intTemp = Integer.parseInt(strCurID);
                                 intTemp--;
                                 strCurID = Integer.toString(intTemp);
                              }
                              //System.out.println("Adding " + strCurDest + "-" + strCurID);
                              // here we need to add the member and initialize its sequence number to -1
                              Vector vTSTemp = new Vector();
                              vTSTemp.addElement(strCurDest);
                              vTSTemp.addElement(strCurID);
                              vCausalMessageIDs.addElement(vTSTemp);
                           }
                        }
                     }
                     else if(strOrdering.equalsIgnoreCase("total"))
                     {
                        // extract the current total sequence number
                        String strTS = stkMessageInfo.nextToken();
                        //System.out.println("The current total sequence is: " + strTS);
                        //strTotalSequence = strTS;
                        cTotalSequence.setCounter(Integer.parseInt(strTS));
                        // clear out the total ordering buffer
                        tOrderer.clearBuffers();
                     }
                  }

                  if(strType.equalsIgnoreCase("sf"))
                  {
                     // here we have been sent a sender fail message--we first need to check if this
                     //  message has ever been received (by searching the history associated with
                     //  the failed sender
                     // first read in the extra header information
                     String strOrigSenderInfo = stkMessageInfo.nextToken();
                     String strOrigMessageID = stkMessageInfo.nextToken();
                     // first let's check the buffer to see if this message has already been
                     //  received and the member is waiting for the gc message from the sender
                     boolean blnBufferFound = false;
                     synchronized (vBuffer)
                     {
                        for(int i=0;i<vBuffer.size();i++)
                        {
                           // extract the current element from the vector
                           Vector vCurBufElem = (Vector) vBuffer.elementAt(i);
                           // now extract the message id and sender
                           String strCurBufSender = (String) vCurBufElem.elementAt(0);
                           String strCurBufMessageID = (String) vCurBufElem.elementAt(1);
                           // check to see if we have found a match
                           if(strOrigSenderInfo.equalsIgnoreCase(strCurBufSender) &&
                                 strOrigMessageID.equalsIgnoreCase(strCurBufMessageID))
                           {
                              // here we have found the matching message in the buffer--
                              //  we can now remove it
                              //synchronized (vBuffer) {
                              vBuffer.removeElement(vCurBufElem);
                              //}
                              // indicate that the element was found
                              blnBufferFound = true;
                              // we also can break out of the for loop
                              break;
                           }
                        }
                     }
                     // check to see if the message was found in the buffer
                     if(blnBufferFound == true)
                     {
                        // in this case the message was found in the buffer, which means that
                        //  this member has already received (and delivered) the message.  Next
                        //  we can make sure not to deliver the message again, load it into the
                        //  buffer, and send an ack message
                        blnDeliverMessage = false;
                        strType = "message";
                     }
                     else
                     {
                        // next set up a loop to find the group member that corresponds to the
                        //  failed sender
                        int i;
                        //synchronized (vGroup) {
                        for(i=0;i<vGroup.size();i++)
                        {
                           // extract the vector that contains the member attributes
                           Vector vMember = (Vector) vGroup.elementAt(i);
                           // next extract the member info (name and listen port)
                           String strCurServer = (String) vMember.elementAt(0);
                           String strCurPort = (String) vMember.elementAt(1);
                           // check to see if we have found a match
                           if(strOrigSenderInfo.equalsIgnoreCase(strCurServer + ":" + strCurPort))
                           {
                              // here we have found a match--next extract the history vector
                              Vector vHistory = (Vector) vMember.elementAt(2);
                              // now we need to try and find the matching entry
                              boolean blnMatchFound = false;
                              for(int j=0;j<vHistory.size();j++)
                              {
                                 String strCurMessageID = (String) vHistory.elementAt(j);
                                 // check for a match
                                 if(strCurMessageID.equalsIgnoreCase(strOrigMessageID))
                                 {
                                    // indicate that the match was found
                                    blnMatchFound = true;
                                    // break out of the loop
                                    break;
                                 }
                              }
                              // check to see if a match was found
                              if(blnMatchFound)
                              {
                                 // here the message has already been received and delivered,
                                 //  so we need to be sure not to deliver it again
                                 blnDeliverMessage = false;
                                 strType = "message";
                              }
                              else
                              {
                                 // here this member never received this message, so we can
                                 //  process it like it was a new message
                                 strType = "message";
                              }
                              break;
                           }
                        }
                        //}
                     }
                     // here we also need to update the group list (remove the faulty sender)
                     removeMember(strOrigSenderInfo);
                  }

                  if(strType.equalsIgnoreCase("message") || strType.equalsIgnoreCase("sequence"))
                  {
                     blnProcessed = true;
                     // here we have been sent a data message--the next line will contain
                     //  the data associated with the message

                     StringTokenizer stkGetMessageData = new StringTokenizer(strRawMessage,"~",true);
                     // get passed the header info
                     stkGetMessageData.nextToken();
                     stkGetMessageData.nextToken();
                     stkGetMessageData.nextToken();
                     stkGetMessageData.nextToken();
                     stkGetMessageData.nextToken();
                     stkGetMessageData.nextToken();
                     stkGetMessageData.nextToken();
                     stkGetMessageData.nextToken();

                     String strMessageData = stkGetMessageData.nextToken();
                     while(stkGetMessageData.hasMoreTokens())
                     {
                        strMessageData = strMessageData + stkGetMessageData.nextToken();
                     }
                     // print out a line to indicate that a message was received
                     //System.out.println("Received a message from " + strSender);
                     //System.out.println("The message was: " + strMessageData);
                     // check to see if the message is from a group member
                     boolean blnIsAMember = findMember(strSender);
                     if(blnIsAMember == true)
                     {
                        // this message is from a group member

                        // check to see if it is a duplicate
                        // next check the active buffer and the histories to see if this message is duplicate
                        boolean blnBufFound = false;
                        synchronized (vBuffer)
                        {
                           for(int i=0;i<vBuffer.size();i++)
                           {
                              // extract the current element from the vector
                              Vector vCurBufElem = (Vector) vBuffer.elementAt(i);
                              // now extract the message id and sender
                              String strCurBufSender = (String) vCurBufElem.elementAt(0);
                              String strCurBufMessageID = (String) vCurBufElem.elementAt(1);
                              // check to see if we have found a match
                              if(strSender.equalsIgnoreCase(strCurBufSender) &&
                                    strMessageID.equalsIgnoreCase(strCurBufMessageID))
                              {
                                 // here we have found the matching message in the buffer--
                                 // indicate that the element was found
                                 blnBufFound = true;
                                 // we also can break out of the for loop
                                 break;
                              }
                           }
                        }
                        // check to see if the message was found in the buffer
                        if(blnBufFound == true)
                        {
                           // this message has already been received and is in the processing buffer--ignore it
                           //System.out.println("Ignoring duplicate message " + strMessageID + " from sender " + strSender);
                           return;
                        }
                        // next we need to check the histories (to make sure the message isn't
                        //  a duplicate that hasn't already been completely processed)
                        //synchronized (vGroup) {
                        for(int i=0;i<vGroup.size();i++)
                        {
                           // extract the vector that contains the member attributes
                           Vector vMember = (Vector) vGroup.elementAt(i);
                           // next extract the member info (name and listen port)
                           String strCurServer = (String) vMember.elementAt(0);
                           String strCurPort = (String) vMember.elementAt(1);
                           // check to see if we have found a match
                           if(strSender.equalsIgnoreCase(strCurServer + ":" + strCurPort))
                           {
                              // here we have found a match--next extract the history vector
                              Vector vHistory = (Vector) vMember.elementAt(2);
                              // now we need to try and find the matching entry
                              boolean blnMatchFound = false;
                              for(int j=0;j<vHistory.size();j++)
                              {
                                 String strCurMessageID = (String) vHistory.elementAt(j);
                                 // check for a match
                                 if(strCurMessageID.equalsIgnoreCase(strMessageID))
                                 {
                                    // indicate that the match was found
                                    blnMatchFound = true;
                                    // break out of the loop
                                    break;
                                 }
                              }
                              // check to see if a match was found
                              if(blnMatchFound)
                              {
                                 // here we have detected that this is a duplicate--ignore it
                                 //System.out.println("Ignoring duplicate message " + strMessageID + " from sender " + strSender);
                                 return;
                              }
                              else
                              {
                                 // break out of the loop
                                 break;
                              }
                           }
                        }
                        //}

                        synchronized (vBuffer)
                        {
                           // next load the message into a vector so that it can be stored in the
                           //  buffer
                           Vector vTemp = new Vector();
                           // add the sender
                           vTemp.addElement(strSender);
                           // add the message id
                           vTemp.addElement(strMessageID);
                           // add the group list (at this point we are ignoring this)
                           vTemp.addElement(strGroupList);
                           // add the message data
                           vTemp.addElement(strMessageData);
                           // now add this vector into the buffer
                           vBuffer.addElement(vTemp);
                        }

                        // next we can check to see if we need to deliver the message
                        if(blnDeliverMessage == true)
                        {
                           // here we can send MessageEvents to all registered listeners
                           if(strType.equalsIgnoreCase("message"))
                           {
                              fireReceiveMessage(strSender,strGroupList,strMessageID,strMessageData);
                           }
                           else
                           {
                              // this is a sequencer message--change what is sent to the total order class
                              //System.out.println("Sending sequencer message to total ordering object");
                              fireReceiveMessage(strSender,strGroupList,strMessageData,"");
                           }
                        }
                        else
                        {
                           // next we may need to update the current message id for ordering
                           if(strOrdering.equalsIgnoreCase("fifo"))
                           {
                              // here we need to find the corresponding record in vFIFOMessageIDs
                              Vector vCurElement = null;
                              int k;
                              boolean blnFound = false;
                              synchronized (vFIFOMessageIDs)
                              {
                                 //synchronized (vFIFOBuffer) {
                                 for(k=0;k<vFIFOMessageIDs.size();k++)
                                 {
                                    // here we need to extract the current vector element
                                    vCurElement = (Vector) vFIFOMessageIDs.elementAt(k);
                                    // extract the sender name
                                    String strCurSender = (String) vCurElement.elementAt(0);
                                    // check to see if we have found a match
                                    if(strCurSender.equalsIgnoreCase(strSender))
                                    {
                                       // here we have found a match--indicate found and break out of the loop
                                       blnFound = true;
                                       break;
                                    }
                                 }

                                 // check to see if the sender was found
                                 if(blnFound == true)
                                 {
                                    // here check to see if this is the next message in the sequence
                                    String strCurMessageID = (String) vCurElement.elementAt(1);
                                    int intCurMessageID = Integer.parseInt(strCurMessageID);
                                    int intMessageID = Integer.parseInt(strMessageID);
                                    if((intMessageID - intCurMessageID) == 1)
                                    {
                                       // here this is the next in the sequence so we need to update
                                       //  vFIFOMessageIDs
                                       // next we need to modify the vFIFOMessageIDs vector element
                                       Vector vTempA = new Vector();
                                       vTempA.addElement(strSender);
                                       vTempA.addElement(strMessageID);
                                       //synchronized (vFIFOMessageIDs) {
                                       vFIFOMessageIDs.setElementAt(vTempA,vFIFOMessageIDs.indexOf(vCurElement));
                                       //}
                                       // next we need to scan the vFIFOBuffer vector to see if there
                                       //  where any messages that where waiting for this message to arrive
                                       //   and be delivered so that they could be delivered
                                       boolean blnChanged = true;
                                       while(blnChanged == true)
                                       {
                                          blnChanged = false;
                                          // set up a loop to go through the vFIFOBuffer elements
                                          for(int j=0;j<vFIFOBuffer.size();j++)
                                          {
                                             Vector vCurFBElement = (Vector) vFIFOBuffer.elementAt(j);
                                             // extract the Sender
                                             String strCurSender = (String) vCurFBElement.elementAt(0);
                                             // check to see if we have found a match
                                             if(strCurSender.equalsIgnoreCase(strSender))
                                             {
                                                // we have found another message from this sender--extract the
                                                //  message id
                                                String strCurFMessageID = (String) vCurFBElement.elementAt(1);
                                                // check to see if this is the next message to be delivered
                                                int intCurFMessageID = Integer.parseInt(strCurFMessageID);
                                                if((intCurFMessageID - intMessageID) == 1)
                                                {
                                                   // here this message needs to be delivered
                                                   MessageEvent mCEvt = (MessageEvent) vCurFBElement.elementAt(2);
                                                   String strDeliver = (String) vCurFBElement.elementAt(3);
                                                   if(strDeliver.equalsIgnoreCase("yes"))
                                                   {
                                                      // make a copy of the listener list in case anyone adds/removes listeners
                                                      Vector targets;
                                                      synchronized (this)
                                                      {
                                                         targets = (Vector) messageListeners.clone();
                                                      }

                                                      // walk through the listener list and call the receiveMessage method in each
                                                      Enumeration e = targets.elements();
                                                      while(e.hasMoreElements())
                                                      {
                                                         MessageListener l = (MessageListener) e.nextElement();
                                                         l.receiveMessage(mCEvt);
                                                      }
                                                   }
                                                   // next we need to modify the vFIFOMessageIDs vector element
                                                   Vector vTempB = new Vector();
                                                   vTempB.addElement(strSender);
                                                   vTempB.addElement(strCurMessageID);
                                                   //synchronized (vFIFOMessageIDs) {
                                                   vFIFOMessageIDs.setElementAt(vTempB,vFIFOMessageIDs.indexOf(vCurElement));
                                                   //}
                                                   // next indicate that a change took place (so a rescan will take place)
                                                   blnChanged = true;
                                                   // reset the value of intMessageID
                                                   intMessageID = intCurFMessageID;
                                                   // remove the element from the buffer
                                                   //synchronized (vFIFOBuffer) {
                                                   vFIFOBuffer.removeElement(vCurFBElement);
                                                   //}
                                                   // next we need to subtract one from the loop variable (to
                                                   //   make sure all elements get searched
                                                   j--;
                                                }
                                             }
                                          }
                                       }
                                    }
                                    else
                                    {
                                       // here we need to add this message to vFIFOBuffer as a message
                                       //  that does not need to be delivered (we are not worrying
                                       //  the ordering on join messages)
                                       Vector vTempA = new Vector();
                                       // add the sender, message id, and mEvt to the vector
                                       vTempA.addElement(strSender);
                                       vTempA.addElement(strMessageID);
                                       MessageEvent bogusME = new MessageEvent(null,strSender,strGroupList,strMessageID,"bogus");
                                       vTempA.addElement(bogusME);
                                       vTempA.addElement("no");
                                       vFIFOBuffer.addElement(vTempA);
                                    }
                                 }
                                 //}
                              }
                           }
                        }

                        // next we can bust up the send info sent to get the server name
                        //  and the port

                        // here we will assume that the message has been delivered (non-uniform)
                        //   so we will log it as received
                        // set up a loop to find the sender in vGroup
                        int i;
                        //synchronized (vGroup) {
                        for(i=0;i<vGroup.size();i++)
                        {
                           // extract the vector that represents the curent member
                           Vector vMember = (Vector) vGroup.elementAt(i);
                           // extract the attributes for the current group member
                           String strCurGServer = (String) vMember.elementAt(0);
                           String strCurGPort = (String) vMember.elementAt(1);
                           // combine these so that a comparison can be made
                           String strCurGMember = strCurGServer + ":" + strCurGPort;
                           // check to see if we have found the matching entry
                           if(strCurGMember.equalsIgnoreCase(strSender))
                           {
                              // here we have found the matching entry--extract the history
                              //  vector
                              Vector vHistory = (Vector) vMember.elementAt(2);
                              // add this message id to the history vector
                              vHistory.addElement(strMessageID);
                              // check the size of the history vector--if it is greater than
                              //  50 then remove the first element
                              if(vHistory.size() > 50)
                              {
                                 vHistory.removeElementAt(0);
                              }
                              Vector vUpdatedMember = (Vector) vMember.clone();
                              // add this vector back into the vMember vector
                              vUpdatedMember.setElementAt(vHistory,2);
                              // add the updated member vector back into the group
                              //synchronized (vGroup) {
                              vGroup.setElementAt(vUpdatedMember,vGroup.indexOf(vMember));
                              //}
                              // break out of the loop
                              break;
                           }
                        }
                        //}

                        StringTokenizer stkSender = new StringTokenizer(strSender,":");
                        String strServerName = stkSender.nextToken();
                        int intDestPort = Integer.parseInt(stkSender.nextToken());

                        // next we can build the header string for the ack message
                        String strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupList + "~" +
                                                  "ack~" + strMessageID;
                        // check to see if we need to use TCP or UDP to send the message
                        if(strConType.equalsIgnoreCase("TCP"))
                        {
                           // next we need to open a socket connection to the sender
                           Socket sock = new Socket(strServerName,intDestPort);
                           byte[] msg = strMessageHeader.getBytes();
                           // now use a DataOutputStream to print the byte array to the socket
                           DataOutputStream dosOut = new DataOutputStream(sock.getOutputStream());
                           dosOut.write(msg);
                           dosOut.flush();
                           dosOut.close();
                           sock.close();
                           /*
                           // now decorate the socket with a print writer
                           PrintWriter pout = new PrintWriter(sock.getOutputStream());
                           // send the ack message back to the sender
                           //System.out.println("Sending ack message back to: " + strSender);
                           pout.println(strMessageHeader);
                           pout.flush();
                           pout.close();
                           sock.close();
                            */
                        }
                        else
                        {
                           // here we need to use UDP to send the message
                           byte[] msg = strMessageHeader.getBytes();
                           // create a new datagram socket to send the message
                           DatagramSocket ds = new DatagramSocket();
                           InetAddress ia = InetAddress.getByName(strServerName);
                           // create the outgoing data packet
                           DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,intDestPort);
                           // now send the packet to the server
                           //System.out.println("Sending ack message back to: " + strSender);
                           ds.send(outPacket);
                           ds.close();
                        }
                        // here we need to kick off a thread to handle if the sender times out
                        //   and never sends the gc message
                        /*
                        MonitorSender ms = new MonitorSender(strSender,strMessageID);
                        Thread tMS = new Thread(ms);
                        // start the thread
                        tMS.start();
                         */
                     }
                     else
                     {
                        // this message has come from a non-member (assume it is a point-to-point message
                        //  from a non-member--here this member will send the message to the group (after it delivers
                        //  it locally)

                        // here we can send MessageEvents to all registered listeners
                        fireReceiveMessage(strSender,strGroupList,strMessageID,strMessageData);
                        sendMessage(strMessageData,"message","","");
                     }
                  }
                  else if(strType.equalsIgnoreCase("ack"))
                  {
                     // here this group member has sent out a message and one of the
                     //  destinations is sending back an ack--first we need to update the
                     //  status vector for this message
                     //System.out.println("Received an ack message from " + strSender);
                     Vector vStatusInfo = new Vector();
                     int i;
                     boolean blnFound = false;
                     Vector vSTemp = null;
                     synchronized (vStatus)
                     {
                        for(i=0;i<vStatus.size();i++)
                        {
                           // first we need to pull the current vector element off of the list
                           vSTemp = (Vector) vStatus.elementAt(i);
                           // next we can extract the first element (which is the message id)
                           String strCurMessageID = (String) vSTemp.elementAt(0);
                           // check to see if we have found a match
                           if(strCurMessageID.equalsIgnoreCase(strMessageID))
                           {
                              // here we have found a match--extract the status vector
                              vStatusInfo = (Vector) vSTemp.elementAt(1);
                              // indicate that the element was found
                              blnFound = true;
                              // break out of the loop
                              break;
                           }
                        }
                        // check to see that the corresponding message was found
                        if(blnFound == true)
                        {
                           blnProcessed = true;
                           // create a boolean variable to indicate whether all acks have been
                           //  received
                           boolean blnAllReceived = true;
                           // next we can update the element that corresponds to this sender
                           for(int j=0;j<vStatusInfo.size();j++)
                           {
                              // extract the current element
                              Vector vSITemp = (Vector) vStatusInfo.elementAt(j);
                              // extract the elements from this vector
                              String strServerName = (String) vSITemp.elementAt(0);
                              String strStatus = (String) vSITemp.elementAt(1);
                              // check to see if the server name matches the sender
                              if(strServerName.equalsIgnoreCase(strSender))
                              {
                                 // here we need to update strStatus to true
                                 strStatus = "true";
                                 // update the status vector
                                 vSITemp.setElementAt(strStatus,1);
                                 vStatusInfo.setElementAt(vSITemp,j);
                              }
                              else
                              {
                                 // check to see if an ack has been received by this member
                                 if(strStatus.equalsIgnoreCase("false"))
                                 {
                                    // indicate that all acks have not been received
                                    blnAllReceived = false;
                                 }
                              }
                           }
                           // next we need to load the statusinfo vector back into status vector
                           Vector vSLoad = new Vector();
                           vSLoad.addElement(strMessageID);
                           vSLoad.addElement(vStatusInfo);
                           //synchronized (vStatus) {
                           vStatus.setElementAt(vSLoad,vStatus.indexOf(vSTemp));
                           //}
                           // next check to see if all acks have been received
                           if(blnAllReceived == true)
                           {
                              //System.out.println("All acks have been received for message: " + strMessageID);
                              // here we need to send the garbage collection message to all members
                              for(int j=0;j<vGroup.size();j++)
                              {
                                 // extract the current group member
                                 Vector vMember = (Vector) vGroup.elementAt(j);
                                 // extract the name and port of the current group member
                                 String strCurServer = (String) vMember.elementAt(0);
                                 String strCurPort = (String) vMember.elementAt(1);
                                 // check to see that this is not the local group member
                                 if(strCurServer.equalsIgnoreCase(strMemberName) && intListenPort == Integer.parseInt(strCurPort))
                                 {
                                    // local member--do nothing
                                 }
                                 else
                                 {
                                    // here we can send the garbage collection message
                                    // next we can build the header string for the gc message
                                    String strMessageHeader = strMemberName + ":" + intListenPort + "~" + strGroupList + "~" +
                                                              "gc~" + strMessageID;
                                    // check to see if we need to use TCP or UDP to send the message
                                    if(strConType.equalsIgnoreCase("TCP"))
                                    {
                                       // next we need to open a socket connection to the sender
                                       Socket sock = new Socket(strCurServer,Integer.parseInt(strCurPort));
                                       byte[] msg = strMessageHeader.getBytes();
                                       // now use a DataOutputStream to print the byte array to the socket
                                       DataOutputStream dosOut = new DataOutputStream(sock.getOutputStream());
                                       dosOut.write(msg);
                                       dosOut.flush();
                                       dosOut.close();
                                       sock.close();
                                       /*
                                       // now decorate the socket with a print writer
                                       PrintWriter pout = new PrintWriter(sock.getOutputStream());
                                       // send the ack message back to the sender
                                       //System.out.println("Sending gc message to: " + strCurServer);
                                       pout.println(strMessageHeader);
                                       pout.flush();
                                       pout.close();
                                       sock.close();
                                        */
                                    }
                                    else
                                    {
                                       // here we need to use UDP to send the message
                                       byte[] msg = strMessageHeader.getBytes();
                                       // create a new datagram socket to send the message
                                       DatagramSocket ds = new DatagramSocket();
                                       InetAddress ia = InetAddress.getByName(strCurServer);
                                       // create the outgoing data packet
                                       DatagramPacket outPacket = new DatagramPacket(msg,msg.length,ia,Integer.parseInt(strCurPort));
                                       // now send the packet to the server
                                       //System.out.println("Sending gc message to: " + strCurServer);
                                       ds.send(outPacket);
                                       ds.close();
                                    }
                                 }
                              }
                              // next we can remove the status vector entry
                              //synchronized(vStatus) {
                              vStatus.removeElement(vSTemp);
                              //}
                           }
                        }
                        else
                        {
                           // here we will ignore the message
                           //System.out.println("Ignoring a message from " + strSender);
                        }
                     }

                  }
                  else if(strType.equalsIgnoreCase("gc"))
                  {
                     // this message is a garbage collection message--we need to remove
                     //  this message from the buffer

                     //System.out.println("Received a garbage collection message from " + strSender);
                     // set up a loop to go through the buffer and find the message with
                     //  matching message id and sender
                     synchronized (vBuffer)
                     {
                        for(int i=0;i<vBuffer.size();i++)
                        {
                           // extract the current element from the vector
                           Vector vCurBufElem = (Vector) vBuffer.elementAt(i);
                           // now extract the message id and sender
                           String strCurBufSender = (String) vCurBufElem.elementAt(0);
                           String strCurBufMessageID = (String) vCurBufElem.elementAt(1);
                           // check to see if we have found a match
                           if(strSender.equalsIgnoreCase(strCurBufSender) &&
                                 strMessageID.equalsIgnoreCase(strCurBufMessageID))
                           {
                              // here we have found the matching message in the buffer--
                              //  we can now remove it
                              //synchronized (vBuffer) {
                              vBuffer.removeElement(vCurBufElem);
                              //}
                              // we also can break out of the for loop
                              blnProcessed = true;
                              break;
                           }
                        }
                     }

                     // check to see if the sender kicked any of the members out
                     if(stkMessageInfo.hasMoreTokens())
                     {
                        // here the next token will be the list of members to be removed
                        String strRemoved = stkMessageInfo.nextToken();
                        //System.out.println("The sender removed the following members: " + strRemoved);
                        // set up a string tokenizer to bust up the removed list and remove the members
                        StringTokenizer stkRemoved = new StringTokenizer(strRemoved,",");
                        // now loop through the entries and remove them
                        while(stkRemoved.hasMoreTokens())
                        {
                           String strToBeRemoved = stkRemoved.nextToken();
                           // call the removeMember method
                           removeMember(strToBeRemoved);
                        }
                     }
                  }
                  if(blnProcessed == true)
                  {
                     // remove the element (since it has been processed)
                     synchronized(vNewMessages)
                     {
                        vNewMessages.removeElement(strTemp);
                     }
                     //itVNM.remove();
                     vNMClone.removeElement(strRawMessage);
                  }
               }

               // put some sleep here
               Thread.sleep(1);
            }
         }
         catch (Exception e)
         {
            e.printStackTrace();
         }

      }

   }

   /** Provide a method to register a listener for MessageEvents */
   synchronized public void addMessageListener(MessageListener l)
   {
      // make sure the listeners vector is initialized
      if(messageListeners == null)
      {
         messageListeners = new Vector();
      }
      // add the new MessageListener to the vector
      messageListeners.addElement(l);
   }

   /** Provide a method to unregister a listener for MessageEvents */
   synchronized public void removeMessageListener(MessageListener l)
   {
      // check to make sure the vector is initialized
      if(messageListeners == null)
      {
         messageListeners = new Vector();
      }
      // remove the element from the vector
      messageListeners.removeElement(l);
   }

   /** Provide a method to fire a MessageEvent to all registered listeners */
   protected void fireReceiveMessage(String strSender, String strGroupList, String strMessageID, String strMessageData)
   {
      // check to see if we have no listeners registered
      if(messageListeners != null)
      {
         // create the event object to send
         MessageEvent mevt = new MessageEvent(this,strSender,strGroupList,strMessageID,strMessageData);
         // here determine if we need to use any ordering scheme (if so--then we
         //  need to fire off another thread to process the message)
         if(strOrdering.equalsIgnoreCase("fifo"))
         {
            // here we need to create a SenderFIFO object and fire off a new thread
            SenderFIFO sndFIFO = new SenderFIFO(mevt);
            // create a new thread to run the SenderFIFO object
            Thread tSF = new Thread(sndFIFO);
            // start the thread to process the message
            tSF.start();
         }
         else if(strOrdering.equalsIgnoreCase("causal"))
         {
            //System.out.println("creating Causal object");
            // here we need to create a Causal object and fire it off in a new thread
            Causal causal = new Causal(mevt);
            // create a new thread to run the Causal object
            Thread tCausal = new Thread(causal);
            // start the thread to process the message
            tCausal.start();
         }
         else if(strOrdering.equalsIgnoreCase("total"))
         {
            //System.out.println("creating Total object");
            tOrderer.processMessage(mevt);
         }
         else
         {
            // here no ordering is required--deliver the messages as the come

            // make a copy of the listener list in case anyone adds/removes listeners
            Vector targets;
            synchronized (this)
            {
               targets = (Vector) messageListeners.clone();
            }

            // walk through the listener list and call the receiveMessage method in each
            Enumeration e = targets.elements();
            while(e.hasMoreElements())
            {
               MessageListener l = (MessageListener) e.nextElement();
               l.receiveMessage(mevt);
            }
         }
      }
   }

   /** define a class to handle implementing Sender-based FIFO ordering
    */
   class SenderFIFO implements Runnable
   {

      private MessageEvent meCurrent;

      /** define a constructor that sets the message object associated with this instance
       */
      public SenderFIFO(MessageEvent mEvt)
      {
         this.meCurrent = mEvt;
      }


      public synchronized void run()
      {
         synchronized (vFIFOMessageIDs)
         {
            synchronized (vFIFOBuffer)
            {
               // first we need to extract the information about this event
               Vector vEventInfo = meCurrent.getMessageData();
               // next extract the Sender and Message ID
               String strSender = (String) vEventInfo.elementAt(0);
               String strMessageID = (String) vEventInfo.elementAt(2);
               //System.out.println("Received message id: " + strMessageID + " in SenderFIFO.");
               // next find the corresponding group member in vFIFOMessageIDs
               boolean blnFound = false;
               Vector vCurElement = null;
               int i;
               for(i=0;i<vFIFOMessageIDs.size();i++)
               {
                  // here we need to extract the current vector element
                  vCurElement = (Vector) vFIFOMessageIDs.elementAt(i);
                  // extract the sender name
                  String strCurSender = (String) vCurElement.elementAt(0);
                  // check to see if we have found a match
                  if(strCurSender.equalsIgnoreCase(strSender))
                  {
                     // here we have found a match--indicate found and break out of the loop
                     blnFound = true;
                     break;
                  }
               }
               // check to see if the sender was found
               if(blnFound == true)
               {
                  // here the sender was found--extract the previous message id
                  String strLastReceivedID = (String) vCurElement.elementAt(1);
                  // check to see if this message can be delivered (is it the next message in the
                  //  sequence)
                  int intLastReceivedID = Integer.parseInt(strLastReceivedID);
                  int intMessageID = Integer.parseInt(strMessageID);
                  if((intMessageID - intLastReceivedID) == 1)
                  {
                     // here we are okay to deliver the message
                     // make a copy of the listener list in case anyone adds/removes listeners
                     Vector targets;
                     synchronized (this)
                     {
                        targets = (Vector) messageListeners.clone();
                     }

                     // walk through the listener list and call the receiveMessage method in each
                     Enumeration e = targets.elements();
                     while(e.hasMoreElements())
                     {
                        MessageListener l = (MessageListener) e.nextElement();
                        l.receiveMessage(meCurrent);
                     }
                     // next we need to modify the vFIFOMessageIDs vector element
                     Vector vTemp = new Vector();
                     vTemp.addElement(strSender);
                     vTemp.addElement(strMessageID);
                     //synchronized (vFIFOMessageIDs) {
                     vFIFOMessageIDs.setElementAt(vTemp,vFIFOMessageIDs.indexOf(vCurElement));
                     //}
                     // next we need to scan the vFIFOBuffer vector to see if there
                     //  where any messages that where waiting for this message to arrive
                     //   and be delivered so that they could be delivered
                     boolean blnChanged = true;
                     while(blnChanged == true)
                     {
                        blnChanged = false;
                        // set up a loop to go through the vFIFOBuffer elements
                        for(int j=0;j<vFIFOBuffer.size();j++)
                        {
                           Vector vCurFBElement = (Vector) vFIFOBuffer.elementAt(j);
                           // extract the Sender
                           String strCurSender = (String) vCurFBElement.elementAt(0);
                           // check to see if we have found a match
                           if(strCurSender.equalsIgnoreCase(strSender))
                           {
                              // we have found another message from this sender--extract the
                              //  message id
                              String strCurMessageID = (String) vCurFBElement.elementAt(1);
                              // check to see if this is the next message to be delivered
                              int intCurMessageID = Integer.parseInt(strCurMessageID);
                              if((intCurMessageID - intMessageID) == 1)
                              {
                                 // here this message needs to be delivered
                                 MessageEvent mCEvt = (MessageEvent) vCurFBElement.elementAt(2);
                                 String strDeliver = (String) vCurFBElement.elementAt(3);
                                 if(strDeliver.equalsIgnoreCase("yes"))
                                 {
                                    // make a copy of the listener list in case anyone adds/removes listeners
                                    synchronized (this)
                                    {
                                       targets = (Vector) messageListeners.clone();
                                    }

                                    // walk through the listener list and call the receiveMessage method in each
                                    e = targets.elements();
                                    while(e.hasMoreElements())
                                    {
                                       MessageListener l = (MessageListener) e.nextElement();
                                       l.receiveMessage(mCEvt);
                                    }
                                 }
                                 // next we need to modify the vFIFOMessageIDs vector element
                                 Vector vTempA = new Vector();
                                 vTempA.addElement(strSender);
                                 vTempA.addElement(strCurMessageID);
                                 //synchronized(vFIFOMessageIDs) {
                                 vFIFOMessageIDs.setElementAt(vTempA,vFIFOMessageIDs.indexOf(vCurElement));
                                 //}
                                 // next indicate that a change took place (so a rescan will take place)
                                 blnChanged = true;
                                 // reset the value of intMessageID
                                 intMessageID = intCurMessageID;
                                 // remove the element from the buffer
                                 //synchronized (vFIFOBuffer) {
                                 vFIFOBuffer.removeElement(vCurFBElement);
                                 //}
                                 // next we need to subtract one from the loop variable (to
                                 //   make sure all elements get searched
                                 j--;
                              }
                           }
                        }
                     }
                  }
                  else
                  {
                     // here this message needs to be loaded into vFIFOBuffer since it
                     //   cannot be delivered yet

                     Vector vTemp = new Vector();
                     // add the sender, message id, and mEvt to the vector
                     vTemp.addElement(strSender);
                     vTemp.addElement(strMessageID);
                     vTemp.addElement(meCurrent);
                     vTemp.addElement("yes");
                     vFIFOBuffer.addElement(vTemp);
                  }
               }
               else
               {
                  // here this member has never sent us a message before--assume this id is the first
                  //  so it is okay to deliver
                  Vector targets;
                  synchronized (this)
                  {
                     targets = (Vector) messageListeners.clone();
                  }

                  // walk through the listener list and call the receiveMessage method in each
                  Enumeration e = targets.elements();
                  while(e.hasMoreElements())
                  {
                     MessageListener l = (MessageListener) e.nextElement();
                     l.receiveMessage(meCurrent);
                  }
                  // next we need add a record to vFIFOMessageIDs
                  Vector vTemp = new Vector();
                  vTemp.addElement(strSender);
                  vTemp.addElement(strMessageID);
                  vFIFOMessageIDs.addElement(vTemp);
               }
            }
         }
      }
   }


   /** define a class to handle implementing Causal ordering (based on vector timestamps)
    */
   class Causal implements Runnable
   {

      private MessageEvent meCurrent;

      /** define a constructor that sets the message object associated with this instance
       */
      public Causal(MessageEvent mEvt)
      {
         this.meCurrent = mEvt;
      }


      public synchronized void run()
      {
         synchronized (vCausalMessageIDs)
         {
            //synchronized (vCausalBuffer) {
            // first we need to extract the information about this event
            Vector vEventInfo = meCurrent.getMessageData();
            // next extract the Sender and Message ID
            String strSender = (String) vEventInfo.elementAt(0);
            String strMessageID = (String) vEventInfo.elementAt(2);
            // here the message ID will be a delimited string--the first element
            //  will be the sequence number from the sender and the rest of
            //  the elements will be the vector timestamp that contains the
            //  causal history for this message
            //System.out.println("Received message id: " + strMessageID + " in Causal.");
            // next we need to bust up the message ID field to extract the vector timestamp
            //  elements
            StringTokenizer stkMessageID = new StringTokenizer(strMessageID,"|,");
            // the first element will contain the message id from the sender (we can
            //  ignore this one for now)
            stkMessageID.nextToken();
            // next we can set up a loop to extract the server names and their
            //  corresponding sequence numbers and determine if it is okay to deliver
            //  this message
            boolean blnOkayToDeliver = true;
            // define a vector to hold the potential update
            Vector vVUpdate = new Vector();
            // define a vector to hold the element that will possibly get updated
            Vector vToBeUpdated = new Vector();
            while(stkMessageID.hasMoreTokens())
            {
               // extract the server info
               String strCurServerInfo = stkMessageID.nextToken();
               // extract the sequence number for this member
               String strCurSequenceNum = stkMessageID.nextToken();
               //System.out.println("Processing the sequence number: " + strCurSequenceNum + " for Member: " + strCurServerInfo);
               // next we need to find the corresponding element in vCausalMessageIDs
               for(int i=0;i<vCausalMessageIDs.size();i++)
               {
                  // first extract the vector element
                  Vector vCurElem = (Vector) vCausalMessageIDs.elementAt(i);
                  // next extract the member info
                  String strCurMemberInfo = (String) vCurElem.elementAt(0);
                  // next extact the current sequence number for this member
                  String strCurLastRecSeq = (String) vCurElem.elementAt(1);
                  if(strCurMemberInfo.equalsIgnoreCase(strCurServerInfo))
                  {
                     // here we have found a match--next we need to check the sequence
                     //  numbers to see if the message can be delivered.  Here if the
                     //  the current member is the sender of the message then the
                     //  sequence number will be one higher than what we have in the
                     //  buffer.  If the current member is not the sender of the message
                     //  then the sequence number should be the same as what we have in
                     //  the buffer
                     //System.out.println("Info in Causal vector is: " + strCurMemberInfo + " - " + strCurLastRecSeq);
                     if(strCurMemberInfo.equalsIgnoreCase(strSender))
                     {
                        // check to see that the sequence number is one larger than what is
                        //  in the buffer
                        int intCurSequenceNum = Integer.parseInt(strCurSequenceNum);
                        int intCurLastRecSeq = Integer.parseInt(strCurLastRecSeq);
                        // first check to see if we have ever received any messages from this user
                        if(intCurLastRecSeq == -1)
                        {
                           // here this is the first message we have received from this user (assume it is okay to
                           //  deliver)
                           //System.out.println("Okay to deliver for " + strCurMemberInfo);
                           // next we need to update the last received message id in the causal vector
                           vVUpdate = new Vector();
                           vVUpdate.addElement(strCurMemberInfo);
                           vVUpdate.addElement(strCurSequenceNum);
                           // next store the element to be updated
                           vToBeUpdated = vCurElem;
                        }
                        else
                        {
                           if((intCurSequenceNum - intCurLastRecSeq) == 1)
                           {
                              //System.out.println("Okay to deliver for " + strCurMemberInfo);
                              // next we need to update the last received message id in the causal vector
                              vVUpdate = new Vector();
                              vVUpdate.addElement(strCurMemberInfo);
                              vVUpdate.addElement(strCurSequenceNum);
                              // next store the element to be updated
                              vToBeUpdated = vCurElem;
                           }
                           else
                           {
                              //System.out.println("Not okay to deliver for " + strCurMemberInfo);
                              blnOkayToDeliver = false;
                           }
                        }
                     }
                     else
                     {
                        // check to see that the sequence number matches the one
                        //  in the buffer (or is less than)
                        int intCurSequenceNum = Integer.parseInt(strCurSequenceNum);
                        int intCurLastRecSeq = Integer.parseInt(strCurLastRecSeq);
                        if((intCurSequenceNum - intCurLastRecSeq) <= 0)
                        {
                           //System.out.println("Okay to deliver for " + strCurMemberInfo);
                        }
                        else
                        {
                           //System.out.println("Not okay to deliver for " + strCurMemberInfo);
                           blnOkayToDeliver = false;
                        }
                     }
                  }
               }
            }

            // check to see if we need to deliver this message or buffer it
            if(blnOkayToDeliver == true)
            {
               // it is okay to deliver this message to all listeners
               Vector targets;
               synchronized (this)
               {
                  targets = (Vector) messageListeners.clone();
               }

               // walk through the listener list and call the receiveMessage method in each
               Enumeration e = targets.elements();
               while(e.hasMoreElements())
               {
                  MessageListener l = (MessageListener) e.nextElement();
                  l.receiveMessage(meCurrent);
               }

               // next we need to update vCausalMessageIDs
               // now replace the element in vCausalMessageIDs
               //synchronized (vCausalMessageIDs) {
               vCausalMessageIDs.setElementAt(vVUpdate,vCausalMessageIDs.indexOf(vToBeUpdated));
               //}

               // next we need to set up a loop to see if any other messages in the buffer
               //  can now be delivered
               boolean blnChanged = true;
               while(blnChanged == true)
               {
                  blnChanged = false;
                  // set up a loop to go through the buffer looking for a message that is
                  //  now ready to deliver
                  for(int j=0;j<vCausalBuffer.size();j++)
                  {
                     Vector vCheckEventInfo = (Vector) vCausalBuffer.elementAt(j);
                     // next extract the Sender and Message ID
                     String strCheckSender = (String) vCheckEventInfo.elementAt(0);
                     String strCheckMessageID = (String) vCheckEventInfo.elementAt(1);
                     // here the message ID will be a delimited string--the first element
                     //  will be the sequence number from the sender and the rest of
                     //  the elements will be the vector timestamp that contains the
                     //  causal history for this message
                     // next we need to bust up the message ID field to extract the vector timestamp
                     //  elements
                     StringTokenizer stkCheckMessageID = new StringTokenizer(strCheckMessageID,"|,");
                     // the first element will contain the message id from the sender (we can
                     //  ignore this one for now)
                     stkCheckMessageID.nextToken();
                     // next we can set up a loop to extract the server names and their
                     //  corresponding sequence numbers and determine if it is okay to deliver
                     //  this message
                     boolean blnCheckOkayToDeliver = true;
                     // define a vector to hold the potential update
                     Vector vCheckVUpdate = new Vector();
                     // define a vector to hold the element that will possibly get updated
                     Vector vCheckToBeUpdated = new Vector();
                     while(stkCheckMessageID.hasMoreTokens())
                     {
                        // extract the server info
                        String strCheckCurServerInfo = stkCheckMessageID.nextToken();
                        // extract the sequence number for this member
                        String strCheckCurSequenceNum = stkCheckMessageID.nextToken();
                        //System.out.println("Processing the sequence number: " + strCheckCurSequenceNum + " for Member: " + strCheckCurServerInfo);
                        // next we need to find the corresponding element in vCausalMessageIDs
                        for(int i=0;i<vCausalMessageIDs.size();i++)
                        {
                           // first extract the vector element
                           Vector vCheckCurElem = (Vector) vCausalMessageIDs.elementAt(i);
                           // next extract the member info
                           String strCheckCurMemberInfo = (String) vCheckCurElem.elementAt(0);
                           // next extact the current sequence number for this member
                           String strCheckCurLastRecSeq = (String) vCheckCurElem.elementAt(1);
                           if(strCheckCurMemberInfo.equalsIgnoreCase(strCheckCurServerInfo))
                           {
                              // here we have found a match--next we need to check the sequence
                              //  numbers to see if the message can be delivered.  Here if the
                              //  the current member is the sender of the message then the
                              //  sequence number will be one higher than what we have in the
                              //  buffer.  If the current member is not the sender of the message
                              //  then the sequence number should be the same as what we have in
                              //  the buffer
                              //System.out.println("Info in Causal vector is: " + strCheckCurMemberInfo + " - " + strCheckCurLastRecSeq);
                              if(strCheckCurMemberInfo.equalsIgnoreCase(strCheckSender))
                              {
                                 // check to see that the sequence number is one larger than what is
                                 //  in the buffer
                                 int intCheckCurSequenceNum = Integer.parseInt(strCheckCurSequenceNum);
                                 int intCheckCurLastRecSeq = Integer.parseInt(strCheckCurLastRecSeq);
                                 // first check to see if we have ever received any messages from this user
                                 if(intCheckCurLastRecSeq == -1)
                                 {
                                    // here this is the first message we have received from this user (assume it is okay to
                                    //  deliver)
                                    //System.out.println("Okay to deliver for " + strCheckCurMemberInfo);
                                    // next we need to update the last received message id in the causal vector
                                    vCheckVUpdate = new Vector();
                                    vCheckVUpdate.addElement(strCheckCurMemberInfo);
                                    vCheckVUpdate.addElement(strCheckCurSequenceNum);
                                    // next store the element to be updated
                                    vCheckToBeUpdated = vCheckCurElem;
                                 }
                                 else
                                 {
                                    if((intCheckCurSequenceNum - intCheckCurLastRecSeq) == 1)
                                    {
                                       //System.out.println("Okay to deliver for " + strCheckCurMemberInfo);
                                       // next we need to update the last received message id in the causal vector
                                       vCheckVUpdate = new Vector();
                                       vCheckVUpdate.addElement(strCheckCurMemberInfo);
                                       vCheckVUpdate.addElement(strCheckCurSequenceNum);
                                       // next store the element to be updated
                                       vCheckToBeUpdated = vCheckCurElem;
                                    }
                                    else
                                    {
                                       //System.out.println("Not okay to deliver for " + strCheckCurMemberInfo);
                                       blnCheckOkayToDeliver = false;
                                    }
                                 }
                              }
                              else
                              {
                                 // check to see that the sequence number matches the one
                                 //  in the buffer
                                 int intCheckCurSequenceNum = Integer.parseInt(strCheckCurSequenceNum);
                                 int intCheckCurLastRecSeq = Integer.parseInt(strCheckCurLastRecSeq);
                                 if((intCheckCurSequenceNum - intCheckCurLastRecSeq) <= 0)
                                 {
                                    //System.out.println("Okay to deliver for " + strCheckCurMemberInfo);
                                 }
                                 else
                                 {
                                    //System.out.println("Not okay to deliver for " + strCheckCurMemberInfo);
                                    blnCheckOkayToDeliver = false;
                                 }
                              }
                           }
                        }
                     }

                     // check to see if we need to deliver this message or buffer it
                     if(blnCheckOkayToDeliver == true)
                     {
                        // it is okay to deliver this message to all listeners
                        blnChanged = true;
                        Vector vChecktargets;
                        synchronized (this)
                        {
                           vChecktargets = (Vector) messageListeners.clone();
                        }

                        // walk through the listener list and call the receiveMessage method in each
                        Enumeration eC = vChecktargets.elements();
                        while(eC.hasMoreElements())
                        {
                           MessageListener l = (MessageListener) eC.nextElement();
                           l.receiveMessage(meCurrent);
                        }

                        // next we need to update vCausalMessageIDs
                        // now replace the element in vCausalMessageIDs
                        //synchronized (vCausalMessageIDs) {
                        vCausalMessageIDs.setElementAt(vCheckVUpdate,vCausalMessageIDs.indexOf(vCheckToBeUpdated));
                        //}
                        vCausalBuffer.removeElement(vCheckEventInfo);
                     }
                  }
               }
            }
            else
            {
               // here we need to store the message in the buffer (it cannot be delivered yet)
               Vector vTemp = new Vector();
               // add the sender, message id, and mEvt to the vector
               vTemp.addElement(strSender);
               vTemp.addElement(strMessageID);
               vTemp.addElement(meCurrent);
               vTemp.addElement("yes");
               vCausalBuffer.addElement(vTemp);
               // next we need to set up a loop to see if any other messages in the buffer
               //  can now be delivered
               boolean blnChanged = true;
               while(blnChanged == true)
               {
                  blnChanged = false;
                  // set up a loop to go through the buffer looking for a message that is
                  //  now ready to deliver
                  for(int j=0;j<vCausalBuffer.size();j++)
                  {
                     Vector vCheckEventInfo = (Vector) vCausalBuffer.elementAt(j);
                     // next extract the Sender and Message ID
                     String strCheckSender = (String) vCheckEventInfo.elementAt(0);
                     String strCheckMessageID = (String) vCheckEventInfo.elementAt(1);
                     // here the message ID will be a delimited string--the first element
                     //  will be the sequence number from the sender and the rest of
                     //  the elements will be the vector timestamp that contains the
                     //  causal history for this message
                     // next we need to bust up the message ID field to extract the vector timestamp
                     //  elements
                     StringTokenizer stkCheckMessageID = new StringTokenizer(strCheckMessageID,"|,");
                     // the first element will contain the message id from the sender (we can
                     //  ignore this one for now)
                     stkCheckMessageID.nextToken();
                     // next we can set up a loop to extract the server names and their
                     //  corresponding sequence numbers and determine if it is okay to deliver
                     //  this message
                     boolean blnCheckOkayToDeliver = true;
                     // define a vector to hold the potential update
                     Vector vCheckVUpdate = new Vector();
                     // define a vector to hold the element that will possibly get updated
                     Vector vCheckToBeUpdated = new Vector();
                     while(stkCheckMessageID.hasMoreTokens())
                     {
                        // extract the server info
                        String strCheckCurServerInfo = stkCheckMessageID.nextToken();
                        // extract the sequence number for this member
                        String strCheckCurSequenceNum = stkCheckMessageID.nextToken();
                        //System.out.println("Processing the sequence number: " + strCheckCurSequenceNum + " for Member: " + strCheckCurServerInfo);
                        // next we need to find the corresponding element in vCausalMessageIDs
                        for(int i=0;i<vCausalMessageIDs.size();i++)
                        {
                           // first extract the vector element
                           Vector vCheckCurElem = (Vector) vCausalMessageIDs.elementAt(i);
                           // next extract the member info
                           String strCheckCurMemberInfo = (String) vCheckCurElem.elementAt(0);
                           // next extact the current sequence number for this member
                           String strCheckCurLastRecSeq = (String) vCheckCurElem.elementAt(1);
                           if(strCheckCurMemberInfo.equalsIgnoreCase(strCheckCurServerInfo))
                           {
                              // here we have found a match--next we need to check the sequence
                              //  numbers to see if the message can be delivered.  Here if the
                              //  the current member is the sender of the message then the
                              //  sequence number will be one higher than what we have in the
                              //  buffer.  If the current member is not the sender of the message
                              //  then the sequence number should be the same as what we have in
                              //  the buffer
                              //System.out.println("Info in Causal vector is: " + strCheckCurMemberInfo + " - " + strCheckCurLastRecSeq);
                              if(strCheckCurMemberInfo.equalsIgnoreCase(strCheckSender))
                              {
                                 // check to see that the sequence number is one larger than what is
                                 //  in the buffer
                                 int intCheckCurSequenceNum = Integer.parseInt(strCheckCurSequenceNum);
                                 int intCheckCurLastRecSeq = Integer.parseInt(strCheckCurLastRecSeq);
                                 // first check to see if we have ever received any messages from this user
                                 if(intCheckCurLastRecSeq == -1)
                                 {
                                    // here this is the first message we have received from this user (assume it is okay to
                                    //  deliver)
                                    //System.out.println("Okay to deliver for " + strCheckCurMemberInfo);
                                    // next we need to update the last received message id in the causal vector
                                    vCheckVUpdate = new Vector();
                                    vCheckVUpdate.addElement(strCheckCurMemberInfo);
                                    vCheckVUpdate.addElement(strCheckCurSequenceNum);
                                    // next store the element to be updated
                                    vCheckToBeUpdated = vCheckCurElem;
                                 }
                                 else
                                 {
                                    if((intCheckCurSequenceNum - intCheckCurLastRecSeq) == 1)
                                    {
                                       //System.out.println("Okay to deliver for " + strCheckCurMemberInfo);
                                       // next we need to update the last received message id in the causal vector
                                       vCheckVUpdate = new Vector();
                                       vCheckVUpdate.addElement(strCheckCurMemberInfo);
                                       vCheckVUpdate.addElement(strCheckCurSequenceNum);
                                       // next store the element to be updated
                                       vCheckToBeUpdated = vCheckCurElem;
                                    }
                                    else
                                    {
                                       //System.out.println("Not okay to deliver for " + strCheckCurMemberInfo);
                                       blnCheckOkayToDeliver = false;
                                    }
                                 }
                              }
                              else
                              {
                                 // check to see that the sequence number matches the one
                                 //  in the buffer
                                 int intCheckCurSequenceNum = Integer.parseInt(strCheckCurSequenceNum);
                                 int intCheckCurLastRecSeq = Integer.parseInt(strCheckCurLastRecSeq);
                                 if((intCheckCurSequenceNum - intCheckCurLastRecSeq) <= 0)
                                 {
                                    //System.out.println("Okay to deliver for " + strCheckCurMemberInfo);
                                 }
                                 else
                                 {
                                    //System.out.println("Not okay to deliver for " + strCheckCurMemberInfo);
                                    blnCheckOkayToDeliver = false;
                                 }
                              }
                           }
                        }
                     }

                     // check to see if we need to deliver this message or buffer it
                     if(blnCheckOkayToDeliver == true)
                     {
                        // it is okay to deliver this message to all listeners
                        blnChanged = true;
                        Vector vChecktargets;
                        synchronized (this)
                        {
                           vChecktargets = (Vector) messageListeners.clone();
                        }

                        // walk through the listener list and call the receiveMessage method in each
                        Enumeration eC = vChecktargets.elements();
                        while(eC.hasMoreElements())
                        {
                           MessageListener l = (MessageListener) eC.nextElement();
                           l.receiveMessage(meCurrent);
                        }

                        // next we need to update vCausalMessageIDs
                        // now replace the element in vCausalMessageIDs
                        //synchronized (vCausalMessageIDs) {
                        vCausalMessageIDs.setElementAt(vCheckVUpdate,vCausalMessageIDs.indexOf(vCheckToBeUpdated));
                        //}
                        vCausalBuffer.removeElement(vCheckEventInfo);
                     }
                  }
               }
            }
         }
         //}
      }
   }

   /** this class returns the total order buffer
    */
   public Vector returnBuffer()
   {
      if(strOrdering.equalsIgnoreCase("total"))
      {
         return null;
      }
      else if(strOrdering.equalsIgnoreCase("causal"))
      {
         return vCausalBuffer;
      }
      else if(strOrdering.equalsIgnoreCase("fifo"))
      {
         return vFIFOBuffer;
      }
      else
      {
         return null;
      }
   }

    public Vector returnMBuffer()
    {
       return vBuffer;
    }
    
	public Vector returnSBuffer()
    {
       return vStatus;
    }

    public void printTMBuffer()
    {
       tOrderer.printMsgBuf();
    }

    public void printTSBuffer()
    {
       tOrderer.printSeqBuf();
    }

    /** define a new class to handle implementing Total ordering
     */
    class newTotal
    {

       private Vector vSequence;
       private Vector vMessage;

       // define an internal class to scan the buffers and deliver messages
       class processBuffers implements Runnable
       {

          public void sendMessage(MessageEvent mEvt)
          {
             if(mEvt != null)
             {
                Vector vChecktargets;
                synchronized (this)
                {
                   vChecktargets = (Vector) messageListeners.clone();
                }

                // walk through the listener list and call the receiveMessage method in each
                Enumeration eC = vChecktargets.elements();
                while(eC.hasMoreElements())
                {
                   MessageListener l = (MessageListener) eC.nextElement();
                   l.receiveMessage(mEvt);
                }
             }
          }

          public void run()
          {
             try
             {
                // this method watches the contents of the buffers and
                //  delivers messages as the can be delivered
                while(true)
                {
                   // clone the contents of vSequence and vMessage
                   Vector vSeqClone = (Vector) vSequence.clone();
                   Vector vMsgClone = (Vector) vMessage.clone();
                   // next loop through the contents of vSeqClone
                   // Iterator itVSC = vSeqClone.iterator();
                   Enumeration itVSC = vSeqClone.elements();
                   while(itVSC.hasMoreElements())
                   {
                      Hashtable htCurSElm = (Hashtable) itVSC.nextElement();
                      // extract the sequence number and see if it is the next to be delivered
                      String strSequence = (String) htCurSElm.get("sequence");
                      int intCurSequence = Integer.parseInt(strSequence);
                      // get the current sequence number
                      int intTotalSequence = cTotalSequence.getCounter();
                      if((intCurSequence - intTotalSequence) == 1)
                      {
                         // this is the next message to be delivered--check to see if this is the sender
                         String strFromSender = (String) htCurSElm.get("fromsender");
                         if(strFromSender.equalsIgnoreCase("yes"))
                         {
                            // here we need to increment the total sequence
                            cTotalSequence.incCounter();
                            // remove the element from the vector
                            vSequence.removeElement(htCurSElm);
                            // itVSC.remove();
                            vSeqClone.removeElement(htCurSElm);
                         }
                         else
                         {
                            String strMessageID = (String) htCurSElm.get("messageID");
                            String strSender = (String) htCurSElm.get("sender");
                            // here we need to scan the message buffer to see if the data has arrived
                            //Iterator itVMC = vMsgClone.iterator();
                            Enumeration itVMC = vMsgClone.elements();
                            while(itVMC.hasMoreElements())
                            {
                               Hashtable htCurMElem = (Hashtable) itVMC.nextElement();
                               String strCurMMessageID = (String) htCurMElem.get("messageID");
                               String strCurMSender = (String) htCurMElem.get("sender");
                               if(strMessageID.equalsIgnoreCase(strCurMMessageID) && strSender.equalsIgnoreCase(strCurMSender))
                               {
                                  // this is the data that goes along with the sequence--send the message
                                  MessageEvent mevt = (MessageEvent) htCurMElem.get("messageEvent");
                                  this.sendMessage(mevt);
                                  // remove the current element from vMessage
                                  vMessage.removeElement(htCurMElem);
                                  // itVMC.remove();
                                  vMsgClone.removeElement(htCurMElem);
                                  // remove the current element from vSequence
                                  vSequence.removeElement(htCurSElm);
                                  //itVSC.remove();
                                  vSeqClone.removeElement(htCurSElm);
                                  // here we need to increment the total sequence
                                  cTotalSequence.incCounter();
                                  // break out of the loop
                                  break;
                               }
                            }
                         }
                      }
                   }
                   // here we have looped through the clones--might need to sleep some here
                   Thread.sleep(1);
                }
             }
             catch (Exception e)
             {
                e.printStackTrace();
             }
          }
       }

       public newTotal()
       {
          vSequence = new Vector();
          vMessage = new Vector();
          // start up a thread to process the buffers
          processBuffers pb = new processBuffers();
          Thread t = new Thread(pb);
          t.start();
       }

       public synchronized void clearBuffers()
       {
          //vSequence.clear();
          vSequence.removeAllElements();
          //vMessage.clear();
          vMessage.removeAllElements();
       }

       public synchronized void CheckSequencer(String strSender,String strMessageID)
       {
          // here we first need to check to see if this member is the sequencer
          //  (the sequencer is the first member in the member group list)
          Vector vSequencerInfo = (Vector) vGroup.elementAt(0);
          // extract the information from the vector
          String strSequencerServer = (String) vSequencerInfo.elementAt(0);
          String strSequencerPort = (String) vSequencerInfo.elementAt(1);
          String strSequencerInfo = strSequencerServer + ":" + strSequencerPort;
          String strMemberInfo = strMemberName + ":" + Integer.toString(intListenPort);
          // check to see if this member is the sequencer
          if(strMemberInfo.equalsIgnoreCase(strSequencerInfo))
          {
             // here this member is the sequencer--we need to send out the sequence message
             //System.out.println("Sending sequence message for: " + strSender + "|" + strMessageID);
             // intSequencer++;
             sendMessage("","sequence",strSender,strMessageID);
          }
       }

       public synchronized void processMessage(MessageEvent meCurrent)
       {
          try
          {
             // first we need to extract the information about this event
             Vector vEventInfo = meCurrent.getMessageData();
             // next extract the Sender and Message ID
             String strSender = (String) vEventInfo.elementAt(0);
             String strMessageID = (String) vEventInfo.elementAt(2);
             //System.out.println("Received message id: " + strMessageID + " in Total.");
             // next we need to check and see if the message is a data message or
             //  a sequence message
             StringTokenizer stkCheckMType = new StringTokenizer(strMessageID,"|");
             // grab the first token and see if it has an "s" on it
             String strCheckTemp = stkCheckMType.nextToken();
             if(strCheckTemp.equalsIgnoreCase("s"))
             {
                // this is a sequence message--the next two elements will be the
                //  sender and message id
                String strOrigSender = stkCheckMType.nextToken();
                String strOrigMessageID = stkCheckMType.nextToken();
                // the last element will be the sequence number from the sequencer
                String strMessageSequenceID = stkCheckMType.nextToken();
                //System.out.println("Received sequence " + strMessageSequenceID +
                //                   " for message " + strOrigSender + "-" + strOrigMessageID);

                // check to see if this member was the original sender
                String strMemberInfo = strMemberName + ":" + Integer.toString(intListenPort);
                Hashtable htNewElem = new Hashtable();
                htNewElem.put("sender",strOrigSender);
                htNewElem.put("messageID",strOrigMessageID);
                if(strOrigSender.equalsIgnoreCase(strMemberInfo))
                {
                   htNewElem.put("fromsender","yes");
                }
                else
                {
                   htNewElem.put("fromsender","no");
                }
                htNewElem.put("sequence",strMessageSequenceID);
                vSequence.addElement(htNewElem);
             }
             else
             {
                // this is a data message
                Hashtable htNewElem = new Hashtable();
                htNewElem.put("sender",strSender);
                htNewElem.put("messageID",strMessageID);
                htNewElem.put("messageEvent",meCurrent);
                htNewElem.put("fromsender","no");
                vMessage.addElement(htNewElem);
                // next we need to check and see if this member is the sequencer
                CheckSequencer(strSender,strMessageID);
             }
          }
          catch (Exception e)
          {
             e.printStackTrace();
             System.exit(-1);
          }
       }

       public void printSeqBuf()
       {
          //Iterator it = vSequence.iterator();
          Enumeration it = vSequence.elements();
          while(it.hasMoreElements())
          {
             Hashtable htCur = (Hashtable) it.nextElement();
             String strSender = (String) htCur.get("sender");
             String strMessageID = (String) htCur.get("messageID");
             String strFromSender = (String) htCur.get("fromsender");
             String strSequence = (String) htCur.get("sequence");
             //System.out.println(strSender + "*" + strMessageID + "*" + strFromSender + "*" + strSequence);
          }
       }

       public void printMsgBuf()
       {
          // Iterator it = vMessage.iterator();
          Enumeration it = vMessage.elements();
          while(it.hasMoreElements())
          {
             Hashtable htCur = (Hashtable) it.nextElement();
             String strSender = (String) htCur.get("sender");
             String strMessageID = (String) htCur.get("messageID");
             String strFromSender = (String) htCur.get("fromsender");
             MessageEvent me = (MessageEvent) htCur.get("messageEvent");
             Vector vMInfo = me.getMessageData();
             String strMessage = (String) vMInfo.elementAt(3);
             System.out.println(strSender + "*" + strMessageID + "*" + strFromSender + "*" + strMessage);
          }
       }

    }

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args)
    {
       try
       {
          // create to members (same system on different ports)
          GroupCom gmA = new GroupCom(10000,"wd26281:10000,wd26281:10001,wd26281:10002","wd26281","UDP");
          GroupCom gmB = new GroupCom(10001,"wd26281:10000,wd26281:10001,wd26281:10002","wd26281","UDP");
          GroupCom gmC = new GroupCom(10002,"wd26281:10000,wd26281:10001,wd26281:10002","wd26281","UDP");
          gmA.startListener();
          gmB.startListener();
          gmC.startListener();
          Thread.sleep(1000);
          gmA.sendMessage("Hello from A","message","","");
          Thread.sleep(2000);
          gmB.sendMessage("Hello from B","message","","");
          Thread.sleep(2000);
          gmC.sendMessage("Hello from C","message","","");
          Thread.sleep(2000);
          gmA.sendMessage("Hello from A again","message","","");
       }
       catch (Exception e)
       {
          e.printStackTrace();
       }

    }
 }
